import numpy as np
import cv2 as cv
import zlib
import base64
from scipy import signal
import json


def image_threshold(image, blur=5, blocksize=11, threshold=None, adaptive=False):
    """Return an binary image. AdaptiveThreshold will keep skeleton of solar modules,
    threshold will only keep contour.

    Parameters
    ----------
    image: array
    Image of solar modules

    blur: int
    Block size in cv.medianBlur

    blocksize: int
    blocksize in cv.adaptiveThreshold

    threshold: int or float
    Threshold value in cv.threshold.
    Pixels above this value will be set 255, and below will be 0.
    If threshold < 1, first threshold (%) of the image grayscale value will be used.
    If not given, first 10% of the image grayscale value will be used.

    adaptive: bool
    if true, adaptiveThreshold will be used # need to change inverse binary into binary

    Returns
    -------
    image_thre: array
    Binary image
    """
    # image_eq = cv.equalizeHist(image_resize)
    image_blur = cv.medianBlur(image, blur)
    if adaptive:
        image_thre = cv.adaptiveThreshold(image_blur, 255, cv.ADAPTIVE_THRESH_GAUSSIAN_C,
                                          cv.THRESH_BINARY, blocksize, 2)
    else:
        if threshold is None:
            threshold = np.quantile(image_blur, 0.1)
        elif threshold < 1:
            threshold = np.quantile(image_blur, threshold)
        _, image_thre = cv.threshold(image_blur, threshold, 255, cv.THRESH_BINARY)

    image_thre = cv.medianBlur(image_thre, 1)

    return image_thre


def base64_2_mask(s):
    """decode the masks from supervisely

    Parameters
    ----------
    s: str
    code of mask from annotated on supervisely

    Returns
    -------
    mask: array of bool
    Image of mask. Only show the convex.
    """
    z = zlib.decompress(base64.b64decode(s))
    n = np.frombuffer(z, np.uint8)
    mask = cv.imdecode(n, cv.IMREAD_UNCHANGED)[:, :, 3].astype(bool)

    return mask


def has_mask(mask_name, path=None, data=None):
    """Check if mask exists
    Parameters
    ----------
    mask_name: str
    The annotation name of the mask.

    path: str or pathlib.PosixPath
    The path of annotation json file

    data: dict
    If provided, will not open path

    Returns
    -------
    If exist, return the index in objects list
    If not, return False
    """
    if path is None and data is None:
        raise ValueError("Mask file not provided.")
    if path:
        with open(path, 'r') as file:
            data = json.load(file)
    
    for inx, obj in enumerate(data["objects"]):
        if obj['classTitle'] == mask_name:
            return inx
    
    return False


def load_mask(path, image, mask_name='module_unet', center=True):
    """Load the image of mask

    Parameters
    ----------
    path: str or pathlib.PosixPath
    The path of annotation json file

    image: array
    The original image of solar module

    mask_name: str
    The annotation name of the mask. Default is 'module_unet'.

    center: bool
    If True, return mask center.

    Returns
    -------
    mask: array
    The image of the mask. The shape is the same as the image of module.

    mask_center: array
    The center coordinate of the mask in the form of [x, y]
    """
    with open(path, 'r') as file:
        data = json.load(file)
    # if len(data["objects"]) == 0:
    #    return None
        # code = data["objects"][0]["bitmap"]["data"]
        # origin = data["objects"][0]["bitmap"]["origin"]
    # else:
    #    flag = True
    #    for obj in data["objects"]:
    #        if obj['classTitle'] == mask_name:
    inx = has_mask(mask_name, data=data)
    if inx is not False:
        obj = data["objects"][inx]
        code = obj["bitmap"]["data"]
        origin = obj["bitmap"]["origin"]
    else:
        mask = np.zeros((image.shape[0], image.shape[1]))
        mask = mask.astype('uint8')
        mask_center = np.array([mask.shape[1]/2, mask.shape[0]/2])
        if center:
            return mask, mask_center
        else:
            return mask
    mask = base64_2_mask(code)
    mask_center = np.array([mask.shape[1]/2, mask.shape[0]/2])
    mask_center += origin

    up = np.zeros((origin[1], mask.shape[1]))
    mask2 = np.vstack((up, mask))
    left = np.zeros((mask2.shape[0], origin[0]))
    mask3 = np.hstack((left, mask2))
    down = np.zeros((image.shape[0] - mask3.shape[0], mask3.shape[1]))
    mask4 = np.vstack((mask3, down))
    right = np.zeros((mask4.shape[0], image.shape[1] - mask4.shape[1]))
    mask5 = np.hstack((mask4, right))

    if center:
        return mask5.astype('uint8'), mask_center.astype(int)
    else:
        return mask5.astype('uint8')


def find_intersection(mask_part, houghlinePara=50):
    """Find the intersection of two edges. Edges detected by houghline method.
    Here we use corner part of a mask to make better detection. (Only one corner need
    to be detected in a corner part).

    Parameters
    ----------
    mask_part: array
    Image of the corner part of a mask

    houghlinePara: int
    Threshold parameter in cv.HoughLines()

    Returns
    -------
    x_cross, y_cross: float
    Coordinate of the intersection
    """
    edge = cv.Canny(mask_part, 0, 1)
    lines = cv.HoughLines(edge, 1, np.pi / 180, houghlinePara)

    rhos = []
    thetas = []
    for line in lines:
        rho, theta = line[0]
        rhos.append(rho)
        thetas.append(theta)

    thetas = np.array(thetas)
    rhos = np.array(rhos)
    mean = np.mean(thetas)
    inx = thetas > mean

    thetas1 = thetas[inx]
    rhos1 = rhos[inx]
    thetas2 = thetas[1 - inx != 0]
    rhos2 = rhos[1 - inx != 0]
    # detect outliers
    inx2 = np.abs(rhos1-np.mean(rhos1)) <= np.std(rhos1)
    rhos1 = rhos1[inx2]
    thetas1 = thetas1[inx2]
    inx3 = np.abs(rhos2-np.mean(rhos2)) <= np.std(rhos2)
    rhos2 = rhos2[inx3]
    thetas2 = thetas2[inx3]

    theta1 = np.mean(thetas1)
    rho1 = np.mean(rhos1)
    theta2 = np.mean(thetas2)
    rho2 = np.mean(rhos2)

    k1 = -1 / np.tan(theta1)
    k2 = -1 / np.tan(theta2)
    b1 = rho1 * np.sin(theta1) - k1 * rho1 * np.cos(theta1)
    b2 = rho2 * np.sin(theta2) - k2 * rho2 * np.cos(theta2)

    x_cross = (b2-b1) / (k1-k2)
    y_cross = (k1 * b2 - k2 * b1) / (k1 - k2)
    # return thetas1, thetas2
    return x_cross, y_cross


def find_module_corner(mask, mask_center, dist=200, displace=0, method=0, corner_center=False, center_displace=10):
    """Detect the corner of solar module. Intersection of edges or corner_detection method is used.

    Parameters
    ----------
    mask: array
    Image of mask

    mask_center: array
    Coordinate of the mask center

    dist: int
    Distance threshold between two corners. Default is 200.

    displace: int
    Displacement of the detected corners to increase tolerance.

    method: int
    0 = use cv.goodFeaturesToTrack() to detect corners
    1 = use cv.HoughLines() to detect edges first and then find the intersections

    corner_center: Bool
    If True, use auto-detected nask center. Otherwise use 'mask_center' parameter. Default is False.

    center_displace: int
    Displacement of the mask center when dividing the mask into four corner parts

    Returns
    -------
    Corners: array
    Sorted coordinates of module corners. The order is top-left, top-right, bottom-left and bottom-right
    """

    x_m = mask_center[0]
    y_m = mask_center[1]

    if corner_center:
        corners = cv.goodFeaturesToTrack(mask, 4, 0.01, 200, blockSize=9)
        corners = np.int0(corners)
        xs1 = []
        ys1 = []
        for i in corners:
            x, y = i.ravel()
            xs1.append(x)
            ys1.append(y)
        x_m = int(np.mean(xs1))
        y_m = int(np.mean(ys1))
    # divide the mask into four corner parts to increase the accuracy
    A = mask[0:y_m+center_displace, 0:x_m+center_displace]
    B = mask[0:y_m+center_displace, x_m-center_displace:]
    C = mask[y_m-center_displace:, 0:x_m+center_displace]
    D = mask[y_m-center_displace:, x_m-center_displace:]

    xs = []
    ys = []
    if method == 0:
        corners_A = cv.goodFeaturesToTrack(A, 1, 0.01, dist, blockSize=9)
        corners_A = np.int0(corners_A)
        corners_B = cv.goodFeaturesToTrack(B, 1, 0.01, dist, blockSize=9)
        corners_B = np.int0(corners_B)
        corners_C = cv.goodFeaturesToTrack(C, 1, 0.01, dist, blockSize=9)
        corners_C = np.int0(corners_C)
        corners_D = cv.goodFeaturesToTrack(D, 1, 0.01, dist, blockSize=9)
        corners_D = np.int0(corners_D)

        for corners in [corners_A, corners_B, corners_C, corners_D]:
            for i in corners:
                x, y = i.ravel()
                xs.append(x)
                ys.append(y)

    if method == 1:
        for part in [A, B, C, D]:
            x_cross, y_cross = find_intersection(part)
            xs.append(x_cross)
            ys.append(y_cross)

    # sort out the corners
    xs[1] += x_m - center_displace
    ys[2] += y_m - center_displace
    xs[3] += x_m - center_displace
    ys[3] += y_m - center_displace

    xs[0] -= displace
    ys[0] -= displace

    xs[1] += displace
    ys[1] -= displace

    xs[2] -= displace
    ys[2] += displace

    xs[3] += displace
    ys[3] += displace

    corners_order = list(zip(xs, ys))

    return np.array(corners_order)


def perspective_transform(image, src, sizex, sizey, rotate=True):
    """Do perspective transform on the solar modules. Orientation of the input module is auto-detected. The output
    module has short side vertically arranged and long side horizontally arranged.

    Parameters
    ----------
    image: array
    Image of solar modules

    src: array
    Sorted coordinates of corners

    sizex, sizey: int
    size of the output image. x is the long side and y is the short side.

    rotate: bool
    If true, auto-detection of orientation is on.

    Returns
    -------
    warped: array
    Transformed image of solar module
    """
    src = np.float32(src)
    
    if rotate and np.sum((src[0] - src[2])**2) > np.sum((src[0] - src[1])**2):
        dst = np.float32([(0, sizey), (0, 0), (sizex, sizey), (sizex, 0)])
    else:
        dst = np.float32([(0, 0), (sizex, 0), (0, sizey), (sizex, sizey)])
    #if np.sum((src[0] - src[2])**2) <= np.sum((src[0] - src[1])**2):
    #    dst = np.float32([(0, 0), (sizex, 0), (0, sizey), (sizex, sizey)])
    #else:
        
    M = cv.getPerspectiveTransform(src, dst)

    warped = cv.warpPerspective(image, M, (sizex, sizey))

    return warped


def find_inner_edge(wrap, dist=25, prom=0.08):  # used to be named as find_cell_corner
    """Detect the inner edges of the transformed solar module. This can be used to verify whether the module
    is successfully transformed.

    Parameters:
    -----------
    wrap: array
    Transformed image of solar module. 
    It should be grayscale or BGR read from Opencv.

    dist: int
    Distance threshold between two edge signals. If the size of solar is large, the threshold should be changed.
    Default is 25 for a 300x600 pixel image of solar module.

    prom: int
    Prominence parameter in scipy.signal.find_peaks(). If signal is weak, the prominence should be decreased.

    Returns
    -------
    peak_x, peak_y: array
    Indices of peaks along x and y directions
    """
    if len(wrap.shape) == 2:
        wrap_g = wrap
    elif len(wrap.shape) == 3:
        wrap_g = cv.cvtColor(wrap, cv.COLOR_BGR2GRAY)

    sum_x = np.sum(wrap_g, axis=0)
    sum_x = sum_x / np.max(sum_x)
    peak_x, _ = signal.find_peaks(-sum_x, distance=dist, prominence=prom)

    sum_y = np.sum(wrap_g, axis=1)
    sum_y = sum_y / np.max(sum_y)
    peak_y, _ = signal.find_peaks(-sum_y, distance=dist, prominence=prom)

    return peak_x, peak_y


# a new version of finding the module corners #
def fill_polygon(points, im_shape):
    """Fill the polygon defined by convex or contour points

    Parameters
    ----------
    points: array
    Coordinates of the points that define the convex or contour of the mask

    im_shape: array
    Array shape of the mask

    Returns
    -------
    im_cnt: array
    Filled contour or convex
    """
    im_cnt = np.zeros((im_shape[0],im_shape[1],1), np.uint8)
    cv.fillPoly(im_cnt, [points], (255,255))

    return im_cnt


def sort_corners(corners):
    """Sort corners for perspective transform

    Parameters
    ----------
    corners: array
    Corners detected by cv.goodFeaturesToTrack()

    Returns
    -------
    sorted_corners: array
    Sorted coordinates of module corners. The order is top-left, top-right, bottom-left and bottom-right.
    """
    col_sorted = corners[np.argsort(corners[:, 1])] # sort on the value in column

    # sort on the value in rows. a, b are the indexes
    a = np.argsort(col_sorted[:2, 0])
    b = np.argsort(col_sorted[2:, 0]) + 2

    return col_sorted[np.hstack((a, b))]


def find_polygon_corners(im, n_corners=4, dist_corners=100, displace=3):
    """Use cv.goodFeaturesToTrack to find the four corners of a filled polygon

    Parameters
    ----------
    im: array
    image of the mask

    n_corners: int
    Number of the corners. Default is 4.

    dist_corners: int
    Distance threshold between two neighbour corners. Default is 100 pixels.

    displace: int
    Displace detected corners to increase tolerance. Default is 3.

    Returns
    -------
    corners_displaced: array
    Detected corners that are sorted and displaced.
    """
    corners_unravel = cv.goodFeaturesToTrack(im, n_corners, 0.01, dist_corners, blockSize=9),
    corners = []
    for corner in corners_unravel[0]:
        corners.append(list(corner.ravel()))
    corners_sorted = sort_corners(np.array(corners))
    corners_displaced = np.array([[-1, -1], [1, -1], [-1, 1], [1, 1]]) * displace + corners_sorted

    return corners_displaced


def find_module_corner2(mask, mode=0):
    """ Detect the corners of a solar module

    Parameters
    ----------
    mask: array
    Image of a mask

    mode: int
    mode == 0: detect corners of the convex of module
    mode == 1: detect corners of the approximated convex of module
    mode == 2: detect corners of the approximated contour of the module
    mode == 3: detect corners of the blurred mask of the module
    mode == 4: detect corners using boudingRect. This is useful if your module is a rounded rectangle.

    Returns
    -------
    Corners: array
    Corners of the solar module
    """
    blur = cv.blur(mask, (6, 6))
    contours, _ = cv.findContours(blur, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)

    inx = 0
    length = 0
    for i, cnt in enumerate(contours):
        if length < len(cnt):
            length = len(cnt)
            inx = i

    if mode == 4:
        rect = cv.minAreaRect(contours[inx])
        corners = cv.boxPoints(rect)
        corners_sorted = sort_corners(np.array(corners))
        corners_displaced = np.array([[-1, -1], [1, -1], [-1, 1], [1, 1]]) * 3 + corners_sorted
        return corners_displaced

    cnt_approx = cv.approxPolyDP(contours[inx], 8, True)
    convex = cv.convexHull(contours[inx])
    conv_approx = cv.approxPolyDP(convex, 8, True)

    if mode == 0:
        im_conv = fill_polygon(convex, blur.shape)
        corners = find_polygon_corners(im_conv)
    elif mode == 1:
        im_conv_app = fill_polygon(conv_approx, blur.shape)
        corners = find_polygon_corners(im_conv_app)
    elif mode == 2:
        im_cnt_app = fill_polygon(cnt_approx, blur.shape)
        corners = find_polygon_corners(im_cnt_app)
    elif mode == 3:
        corners = find_polygon_corners(blur)
    else:
        raise Exception("mode must be one of 0, 1, 2, 3, 4")
    
    return corners




import cv2 as cv
import numpy as np
import matplotlib.pyplot as plt


def lightness_balance(src, threshold=[1, 99], extend=[0.1, 0.9]):
    # compute the lightness of source image
    hsv_image = cv.cvtColor(src, cv.COLOR_BGR2HSV)
    lightness = hsv_image[:, :, 2].mean()

    # set a threshold of min_grayscale and max_grayscale,
    # outside which are removed
    max_percentile_pixel = np.percentile(src, threshold[1])
    min_percentile_pixel = np.percentile(src, threshold[0])

    # remove the pixels of grayscale outside the threshold
    src[src >= max_percentile_pixel] = max_percentile_pixel
    src[src <= min_percentile_pixel] = min_percentile_pixel

    # balance brightness
    out = np.zeros(src.shape, src.dtype)
    cv.normalize(src, out, 255*extend[0], 255*extend[1], cv.NORM_MINMAX)

    return out


def edge_remove(im_src, threshold=0.2878, displace=3, size=400):
    # find the edge, here only find horizontal edge
    image = im_src / np.max(im_src)
    image_mask = image > threshold
    sum_1 = np.sum(image_mask, axis=1)
    sum_1 = sum_1 / np.max(sum_1)
    
    inx = np.argwhere(sum_1 > 0)
    top = inx[5]-5
    bottom = inx[-5]+5
    if top > displace:
        top -= displace
    else:
        top = 0
    if bottom < (im_src.shape[0]-displace):
        bottom += displace
    else:
        bottom = im_src.shape[-1]
    top = int(top)
    bottom = int(bottom)

    # transform
    src = [(0, top), (im_src.shape[-1], top), (0, bottom),
           (im_src.shape[-1], bottom)]
    src = np.float32(src)
    dst = [(0, 0), (size, 0), (0, size), (size, size)]
    dst = np.float32(dst)
    M = cv.getPerspectiveTransform(src, dst)
    
    warped = cv.warpPerspective(im_src, M, (size, size))
    
    return warped




import numpy as np
import cv2 as cv
import json
from pathlib import Path
from scipy.signal import find_peaks
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
from copy import deepcopy


def linear(x, a, b):
    """Linear function
    Parameters
    ----------
    x, a, b: float
    linear function f = ax + b

    Returns
    -------
    f: float
    linear function f = ax + b
    """
    return a * x + b


def linear_regression(inxes, lines, outlier_filter=False, threshold=0.3):
    """Linear regression: f = ax + b. Return multiple fitted lines. After detect_edge, do linear fitting to
    return the final parameters(a and b) of edges.

    Parameters
    ----------
    inxes: list, or array
    Values of x

    lines: list, or array
    Each line in lines represents values of f.

    outlier_filter: bool
    If true, outliers will be filtered

    threshold: float
    Filter threshold

    Returns
    -------
    ab_s: list
    List of a and b of multiple lines
    """
    ab_s = []

    for line in lines:
        ab, _ = curve_fit(linear, inxes, line)
        if outlier_filter:
            line_fit = ab[0] * inxes + ab[1]
            mse = np.mean((line - line_fit) ** 2)
            cook_distances = []
            outlier_inx = []
            for inx, element in enumerate(line):
                inx_i = np.delete(inxes, inx)
                line_i = np.delete(line, inx)
                ab_i, _ = curve_fit(linear, inx_i, line_i)
                line_fit_i = ab_i[0] * inxes + ab_i[1]
                cook_distance = (np.sum((line_fit - line_fit_i) ** 2)) / mse
                cook_distances.append(cook_distance)
                if cook_distance > threshold:
                    outlier_inx.append(inx)
            inxes_new = np.delete(inxes, outlier_inx)
            line_new = np.delete(line, outlier_inx)
            ab, _ = curve_fit(linear, inxes_new, line_new)

        ab_s.append(ab)

    ab_s = np.array(ab_s)

    return ab_s


def detectoutliers(data, rate=1.5, option=1):
    # may be replaced by a better algorithm
    """ Detect outliers among peaks(internal edges) of a split.
    If option == 0, will replace outliers.
    If option == 1, will remove outliers. (e.g. some extra wrong peaks)

    Parameters
    ----------
    data: array
    Peaks of internal edges

    rate: float
    Tolerance of good data

    option: int
    """

    bottom = np.percentile(data, 25)
    up = np.percentile(data, 75)
    IQR = up - bottom
    outlier_step = rate * IQR

    outlier_list = (data < bottom - outlier_step) | (data > up + outlier_step)
    if option == 0:
        for inx, flag in enumerate(outlier_list):
            if flag:
                left = inx
                right = inx
                while outlier_list[left]:
                    left -= 1
                    if left <= 0:
                        left = inx
                        while outlier_list[left]:
                            left += 1

                while outlier_list[right]:
                    right += 1
                    if right >= len(outlier_list) - 1:
                        right = inx
                        while outlier_list[right]:
                            right -= 1
                outlier_list[inx] = False
                data[inx] = (data[left] + data[right]) / 2

        return data

    if option == 1:
        outlier_list = np.insert(outlier_list, 4, False)
        for i in range(5, len(outlier_list) - 5):
            if outlier_list[i]:
                outlier_list[i] = False

        return (1 - outlier_list).astype(np.bool)


def filter_margin(edges, im_length, margin=20):
    """Filter the wrongly detected edges on the margin of the image

    Parameters
    ----------
    edges: list or array
    Detected positions of edges

    im_length: int
    Length of width or height of the image. If detect vertical edges, use width.
    If detect horizontal edges, use height

    margin: int
    Margin in which there shouldn't be edges but may be wrongly detected. Default is 20 pixels

    Returns
    -------
    edges: list or array
    Filtered edges
    """
    if edges[0] < margin:
        edges = np.delete(edges, 0)
    if edges[-1] > (im_length - margin):
        edges = np.delete(edges, -1)

    return edges


def split_img(image, n_split=None, split_size=None, direction=0):
    """Split the image along x/y axis.

    Parameters
    ----------
    image: array
    Image to be split

    n_split: int
    Number of splits.

    split_size: int
    Size of each split. Ignored if SPLIT is provided.

    direction: int
    If 0: along y axis. Generate horizontal splits. Detect vertical lines.
    If 1: along x axis. Generate vertical splits. Detect horizontal lines.

    Returns
    -------
    splits
    """
    if (direction != 1) and (direction != 0):
        raise ValueError("Direction should be either 0 or 1.")

    dimension = image.shape[direction]
    if split_size and (n_split is None):
        n_split = int(dimension / split_size)
    elif (split_size is None) and (n_split is None):
        raise ValueError("Either split or split_size should be provided.")

    end = int(dimension / n_split)
    if direction == 0:
        splits = np.vsplit(image[: end * n_split, :], n_split)
        if end * n_split < dimension:
            splits.append(image[end * n_split:, :])
    elif direction == 1:
        splits = np.hsplit(image[:, :end * n_split], n_split)
        if end * n_split < dimension:
            splits.append(image[:, end * n_split:])

    return splits


def detect_peaks(split, direction, busbar, thre=0.9, cell_size=None, interval=None, margin=None):
    """Detect peaks which correspond to internal edges

    Parameters
    ----------
    split: array
    Split of the image

    direction: int
    If 0: along y axis. Process horizontal splits. Detect vertical lines.
    If 1: along x axis. Process vertical splits. Detect horizontal lines.

    cell_size: int
    Size (pixel) of a cell in the raw image. If provided, peak interval can be
    automatically computed

    busbar: int
    Number of busbars of a cell.

    thre: float
    Peak intensity above THRE will be set as 1.
    Note that the edge's peak intensity should be lowest because edges are black

    interval: int
    Distance between each peak.
    If direction == 0, default is int(cell_size * 0.95)
    If direction == 1, Default is int(cell_size / (busbar + 1) * 0.5)

    margin: int
    Margin in which there shouldn't be edges but may be wrongly detected.
    """
    if (direction != 1) and (direction != 0):
        raise ValueError("Direction should be either 0 or 1.")

    if cell_size is None and interval is None:
        raise ValueError("Either cell_size or interval value should be provided.")

    if interval is None and direction == 0:
        interval = int(cell_size * 0.95)
    elif interval is None and direction == 1:
        interval = int(cell_size / (busbar + 1) * 0.5)

    sum_split = np.sum(split, axis=direction)
    sum_split = sum_split / np.max(sum_split)
    sum_split_thre = deepcopy(sum_split)
    sum_split_thre[sum_split > thre] = 1

    peaks, _ = find_peaks(-1 * sum_split_thre, distance=interval)
    if margin:
        peaks = filter_margin(peaks, im_length=split.shape[direction - 1], margin=margin)

    return sum_split, sum_split_thre, peaks


def plot_peaks_one(n, image, busbar, n_split=None, split_size=None, direction=0,
                   thre=0.9, cell_size=None, interval=None, margin=None):
    splits = split_img(image, n_split, split_size, direction)
    split = splits[n]
    sum_split, sum_split_thre, peaks = detect_peaks(split=split, direction=direction, busbar=busbar, thre=thre,
                                                    cell_size=cell_size, interval=interval, margin=margin)

    plt.subplot(311)
    plt.imshow(split, 'gray')
    plt.subplot(312)
    plt.plot(list(range(len(sum_split))), sum_split)
    plt.scatter(peaks, sum_split[peaks])
    plt.subplot(313)
    plt.plot(list(range(len(sum_split_thre))), sum_split_thre)
    plt.scatter(peaks, sum_split_thre[peaks])
    plt.tight_layout()


def plot_peaks_three(n_list, image, busbar, n_split=None, split_size=None, direction=0,
                     thre=0.9, cell_size=None, interval=None, margin=None):
    splits = split_img(image, n_split, split_size, direction)
    for i, n in enumerate(n_list):
        split = splits[n]
        sum_split, sum_split_thre, peaks = detect_peaks(split=split, direction=direction, busbar=busbar, thre=thre,
                                                        cell_size=cell_size, interval=interval, margin=margin)

        plt.subplot(330 + i * 3 + 1)
        plt.imshow(split, 'gray')
        if i == 0:
            plt.title('Module split')
        plt.subplot(330 + i * 3 + 2)
        plt.plot(list(range(len(sum_split))), sum_split)
        plt.scatter(peaks, sum_split[peaks])
        if i == 0:
            plt.title("Original sum")
        plt.subplot(330 + i * 3 + 3)
        plt.plot(list(range(len(sum_split_thre))), sum_split_thre)
        plt.scatter(peaks, sum_split_thre[peaks])
        if i == 0:
            plt.title("Threshold sum")
    plt.tight_layout()


def detect_vertical_lines(image_thre, column, cell_size, thre=0.8, split=100, peak_interval=None, margin=None):
    """ Detect vertical edges by segmenting image into horizontal splits. Note that the module
    is not perpective transformed yet.

    Parameters
    ----------
    image_thre: array
    Adaptive threshold of raw images

    column: int
    Number of columns of solar module

    cell_size: int
    Output cell size in pixel

    thre: float
    Peak intensity above THRE will be set as 1.
    Note that the edge's peak intensity should be lowest because edges are black

    split: int
    Number of splits

    peak_interval: int
    Distance between each peak. Default is int(cell_size * 0.95)

    Returns
    -------
    vline_abs_couple: array
    Suppose a line is y=a*x+b.
    Return a and b of a couple edges (left and right of a cell).
    """
    # height = image_thre.shape[0]
    # end = int(height / split)
    # image_hsplits = np.vsplit(image_thre[: end * split, :], split)  # horizontal splits
    # image_hsplits.append(image_thre[end * split:, :])
    image_hsplits = split_img(image_thre, split=split, direction=0)

    edge_x = []
    inx_y = []
    for inx, im_split in enumerate(image_hsplits):
        # sum_split = np.sum(im_split, axis=0)
        # sum_split = sum_split / np.max(sum_split)
        # sum_split[sum_split > thre] = 1

        # if peak_interval is None:
        #    peak_interval = int(cell_size * 0.95)
        # peak, _ = find_peaks(-1 * sum_split, distance=peak_interval)
        _, _, peak = detect_peaks(im_split, direction=0, cell_size=cell_size,
                                  busbar=None, thre=thre, interval=peak_interval, margin=margin)
        if len(peak) > column - 2:
            peak_new = [peak[0]]
            for i in range(1, len(peak) - 1):
                if np.abs(peak[i] - peak[i + 1]) < 15:
                    peak_mean = (peak[i] + peak[i + 1]) / 2
                    peak_new.append(peak_mean)
                elif np.abs(peak[i] - peak[i - 1]) > 15:
                    peak_new.append(peak[i])

            peak_new.append(peak[-1])
            peak_new = np.array(peak_new)
            peak_new_a = np.delete(peak_new, 0)
            peak_new_b = np.delete(peak_new, -1)
            peak_new_detect = peak_new[detectoutliers(np.abs(peak_new_a - peak_new_b), option=1)]

            if len(peak_new_detect) == 1 + column:
                edge_x.append(peak_new_detect)
                inx_mean = ((2 * inx + 1) * (image_thre.shape[0] / split) - 1) / 2
                inx_y.append(inx_mean)
    edge_x = np.array(edge_x)

    vlines = list(zip(*edge_x))  # line parallel to y axis
    vlines = np.array(vlines)
    inx_y = np.array(inx_y)
    # for lines in vlines:
    #    lines_new = self.detectoutliers(lines, option=0)
    #    while np.std(lines_new) > 15:
    #        lines_new = self.detectoutliers(lines, rate=1, option=0)

    # v_abs = []
    # for verticaline in vlines:
    #    ab, _ = curve_fit(self.linear, inx_y, verticaline) # x = ay + b
    #    v_abs.append(ab)
    v_abs = linear_regression(inx_y, vlines, outlier_filter=True)
    # temp1 = v_abs.copy()
    temp1 = np.delete(v_abs, -1, 0)
    # temp2 = v_abs.copy()
    temp2 = np.delete(v_abs, 0, 0)

    vline_abs_couple = np.array(list(zip(temp1, temp2)))
    # vline_abs = [(v_abs[i],v_abs[i+1]) for i in range(0, len(v_abs), 2)] # put edges of each cell into a tuple

    return vline_abs_couple


def detect_horizon_lines(image_thre, row, busbar, cell_size, thre=0.6, split=50, peak_interval=None, margin=None):
    """ Detect horizontal edges by segmenting image into vertical splits. Note that the module
    is not perpective transformed yet.

    Parameters
    ---------
    image_thre: array
    Adaptive threshold of raw images

    row: int
    Number of rows of solar module

    busbar: int
    Number of busbars of a solar cell

    cell_size: int
    Output cell size in pixel

    thre: float
    Peak intensity above THRE will be set as 1.
    Note that the edge's peak intensity should be lowest because edges are black

    split: int
    Number of splits

    peak_interval: int
    Distance between each peak.

    Returns
    -------
    hline_abs_couple: array
    Suppose a line is y=a*x+b.
    Return 'a' and 'b' of a couple edges (top and bottom of a cell).
        """
    # width = image_thre.shape[1]
    # end = int(width / split)
    # image_vsplits = np.hsplit(image_thre[:, :end * split], split)  # vertical splits
    # image_vsplits.append(image_thre[:, end * split:])
    image_vsplits = split_img(image_thre, split=split, direction=1)

    edge_y = []
    inx_x = []
    for inx, im_split in enumerate(image_vsplits):
        # sum_split = np.sum(im_split, axis=1)
        # sum_split = sum_split / np.max(sum_split)
        # sum_split[sum_split > thre] = 1

        # if peak_interval is None:
        #    peak_interval = int(cell_size / (busbar + 1) * 0.5)
        # peak, _ = find_peaks(-1 * sum_split, distance=peak_interval)
        _, _, peak = detect_peaks(im_split, 1, cell_size, busbar, thre, peak_interval, margin=margin)
        if len(peak) >= row * (busbar + 1) - 1:
            peak_new = [peak[0]]
            for i in range(1, len(peak) - 1):
                if np.abs(peak[i] - peak[i + 1]) < 15:
                    peak_mean = (peak[i] + peak[i + 1]) / 2
                    peak_new.append(peak_mean)
                elif np.abs(peak[i] - peak[i - 1]) > 15:
                    peak_new.append(peak[i])

            peak_new.append(peak[-1])
            peak_new = np.array(peak_new)
            peak_new_a = np.delete(peak_new, 0)
            peak_new_b = np.delete(peak_new, -1)
            peak_new_detect = peak_new[detectoutliers(np.abs(peak_new_a - peak_new_b), rate=0.5, option=1)]

            if len(peak_new_detect) == (busbar + 1) * row + 1:
                edge_y.append(peak_new_detect)
                inx_mean = ((2 * inx + 1) * (image_thre.shape[1] / split) - 1) / 2
                inx_x.append(inx_mean)

    edge_y = np.array(edge_y)

    hlines = list(zip(*edge_y))
    hlines = np.array(hlines)
    inx_x = np.array(inx_x)
    # for lines in hlines:
    #    lines_new = self.detectoutliers(lines, option=0)
    #    while np.std(lines_new) > 10:
    #        lines_new = self.detectoutliers(lines, rate=1, option=0)

    # hb_abs = [] # all lines including busbar
    hb_abs = linear_regression(inx_x, hlines, outlier_filter=True)
    hline_abs_couple = []  # all lines excluding busbar

    # for horizonline in hlines:
    #    ab, _ = curve_fit(self.linear, inx_x, horizonline) # y = ax + b
    #    hb_abs.append(ab)

    hline_abs_couple = [(hb_abs[(busbar + 1) * i], hb_abs[(busbar + 1) * (i + 1)]) for i in range(row)]
    # hline_abs = [(hb_abs[(4+1)*i],hb_abs[(4+1)*(i+1)]) for i in range(6)]
    # hline_abs = [(hb_abs[(self.busbar+2)*i],hb_abs[(self.busbar+2)*(i+1)-1]) for i in range(self.row)]

    return hline_abs_couple


def detect_edge(image, row_col, busbar, peaks_on=0, split_size=10, thre=0.9,
                cell_size=None, peak_interval=None, margin=20):
    """Detect the inner edges of a solar module. Split the solar module first, then detect position of
    edges in each split.
    Note that the solar module is already perspectively transformed.

    Parameters
    ----------
    image: array
    Perspective transformed image of solar module. 
    Input should be grayscale or BGR read from Opencv

    row_col: list
    Number of rows/columns of solar module

    cell_size: int
    Estimated size of single cells in pixels. Default is 30.

    busbar: int
    number of busbar

    peaks_on: int
    Detect horizontal edges or vertical edges
    If 0: detect vertical edges
    If 1: detect horizontal edges

    split_size: int
    The size of each split

    thre: float
    Peak intensity above THRE will be set as 1.
    Note that the edge's peak intensity should be lowest because edges are black

    peak_interval: int
    Distance between each peak.

    margin: int
    Margin in which there shouldn't be edges but may be wrongly detected. Default is 20 pixels

    Returns
    -------
    split_inx: array
    Index of each split

    edge_list: array
    Position of edges of each split.
    Form is [[Positions of first edge in each split], [Positions of second edges in each split], ...]
    """
    if len(image.shape) == 2:
        image_g = image
    elif len(image.shape) == 3:
        image_g = cv.cvtColor(image, cv.COLOR_BGR2GRAY)

    if peaks_on != 0 and peaks_on != 1:
        raise ValueError('peaks_on must be 0 or 1')

    splits = split_img(image_g, split_size=split_size, direction=peaks_on)

    peaklist = []
    splits_inx = []

    _, _, peaks_whole = detect_peaks(split=image_g, direction=peaks_on, busbar=busbar, thre=thre,
                                     cell_size=cell_size, interval=peak_interval, margin=margin)

    flag = False
    for inx, split in enumerate(splits):
        _, _, peaks_split = detect_peaks(split=split, direction=peaks_on, busbar=busbar,
                                         cell_size=cell_size, interval=peak_interval, margin=margin)

        splits_inx.append(int(split_size * (inx + 1 / 2)))
        if len(peaks_split) == row_col[peaks_on - 1] - 1:
            peaklist.append(peaks_split)
        elif (len(peaks_split) != row_col[peaks_on - 1] - 1) and inx != 0:
            peaklist.append(peaklist[inx - 1])
        elif len(peaks_split) == row_col[peaks_on - 1] - 1:
            peaklist.append(peaks_whole)
        else:
            peaklist.append('nan')
            flag = True

    # use any other splits to represent split[0, 1, 2...] if peaklist[0] fails
    if flag:
        peaklist = np.array(peaklist)
        nan = np.argwhere(peaklist == 'nan')
        n_nan = np.argwhere(peaklist != 'nan')
        for i in nan:
            peaklist[int(i)] = peaklist[int(n_nan[0])]
        peaklist = list(peaklist)

    edgelist = np.array(list(zip(*peaklist)))
    edgelist_c = np.copy(edgelist)

    if len(peaks_whole) == row_col[peaks_on - 1] - 1:
        for i, edge in enumerate(edgelist_c):
            for j, sub_edge in enumerate(edge):
                if np.abs(sub_edge - peaks_whole[i]) > 10:
                    edgelist_c[i][j] = peaks_whole[i]

    return np.array(splits_inx), edgelist_c


def displace(line_ab, displacement):
    """Displace fitted edges to increase tolerance

    Parameters
    ----------
    line_ab: list
    [a, b] of fitted edges

    displacement: int
    Perpendicular distance of moving the lines

    Returns
    (a, b): tuple
    a, b of displaced line
    """
    c = displacement * np.sqrt((1 + line_ab[0] ** 2))

    return (line_ab[0], line_ab[1] + c)


def couple_edges(lines_ab, length, displacement=0, add_edge=True):
    """Couple two neighbour edges

    Parameters
    ----------
    lines_ab: list or array
    [a, b] of fitted edges

    length: int
    Width or height of the image.
    If couple vertical edges, use width.
    If couple horizontal edges, use height

    displacement: int
    Displacement of edges to increase tolerance
    The first line in the couple moves to negative direction, the second to positive direction

    add_edge: bool
    If true, add the initial and tail edges, which are the two ends of images

    Returns
    -------
    lines_couple: array
    Coupled edges
    """
    #
    lines_copy = np.copy(lines_ab)
    if add_edge:
        lines_copy = np.insert(lines_copy, 0, [0, 0], axis=0)
        lines_copy = np.insert(lines_copy, len(lines_copy), [0, length - 1], axis=0)
    lines_l = np.delete(lines_copy, -1, 0)
    lines_r = np.delete(lines_copy, 0, 0)
    lines_couple = np.array(list(zip(lines_l, lines_r)))

    lines_couple_new = []
    for couple in lines_couple:
        minus = displace(couple[0], -displacement)
        plus = displace(couple[1], displacement)
        lines_couple_new.append((minus, plus))

    return np.array(lines_couple_new)


def correction(lines, displacement):
    """The first line in the tuple moves to negative direction,
    the second to positive direction.

    Parameters
    ----------
    lines: list or array

    """
    lines_new = []
    for couple in lines:
        minus = displace(couple[0], -displacement)
        plus = displace(couple[1], displacement)
        lines_new.append((minus, plus))

    return lines_new


def cross(vline, hline):
    """Intersction of edges

    Parameters
    ----------
    vline, hline: list or tuple
    (a, b) of vertical edge, horizontal edge
    """
    x = (vline[0] * hline[1] + vline[1]) / (1 - vline[0] * hline[0])
    y = (hline[0] * vline[1] + hline[1]) / (1 - vline[0] * hline[0])

    return (x, y)


def perspective_transform(image, src, size):
    """Crop out a solar cell

    Parameters
    ----------
    image: array
    Perspective transformed image of solar module

    src: array
    Sorted coordinates of corners. [top-left, top-right, bottom-left, bottom-right]

    size: int
    Size of cropped solar cell, e.g., to output 32x32 cells, input 32.

    Returns
    -------
    warped: array
    Transformed solar cell image
    """
    src = np.float32(src)
    dst = np.float32([(0, 0), (size, 0), (0, size), (size, size)])
    M = cv.getPerspectiveTransform(src, dst)

    warped = cv.warpPerspective(image, M, (size, size))

    return warped


def segment_cell(image, hline_abs_couple, vline_abs_couple, cellsize,
                 save_path=None, displacement=None):
    """Crop out solar cells from a module

    Parameters
    ----------
    image: array
    Perspective transformed image of solar module

    hline_abs_couple, vline_abs_couple,: array
    Couple of horizontal edges, vertical edges

    cellsize: int
    Size of cropped solar cell, e.g., to output 32x32 cells, input 32.

    save_path: str or pathlib.PosixPath
    Path to store the output

    displacement
    If displace, then move edge couples with displacement

    Returns
    -------
    cells: array
    Images of cropped cells
    """
    if displacement:
        hline_abs_couple = correction(hline_abs_couple, displacement)
        vline_abs_couple = correction(vline_abs_couple, displacement)

    cells = []
    for i, hline_ab in enumerate(hline_abs_couple):
        for j, vline_ab in enumerate(vline_abs_couple):
            xy = []
            for hab in hline_ab:
                for vab in vline_ab:
                    xy.append(cross(vab, hab))

            warped = perspective_transform(image, xy, cellsize)
            cells.append(warped)

            # counter += 1

            if save_path:
                cv.imwrite(str(save_path / (str(i) + str(j) + '.png')), warped)

    return cells


def plot_edges(image, hline_abs_couple, vline_abs_couple, displacement=1, linewidth=1):
    """Plot the detected edges

    Parameters
    ----------
    image: array
    Raw image

    hline_abs_couple, vline_abs_couple: list or array
    a and b of edge couples

    displacement: int
    Displace detected edges

    linewidth: int
    Width of plotted lines
    """
    newhline_abs = correction(hline_abs_couple, displacement)
    newvline_abs = correction(vline_abs_couple, displacement)

    plt.imshow(image, cmap='gray')
    x = np.array(range(image.shape[1]))
    y = np.array(range(image.shape[0]))

    for h_ab in newhline_abs:
        plt.plot(x, h_ab[0][0] * x + h_ab[0][1], linewidth=linewidth, linestyle='--', color='r')
        plt.plot(x, h_ab[1][0] * x + h_ab[1][1], linewidth=linewidth, linestyle='--', color='r')

    for v_ab in newvline_abs:
        plt.plot(v_ab[0][0] * y + v_ab[0][1], y, linewidth=linewidth, linestyle='--', color='r')
        plt.plot(v_ab[1][0] * y + v_ab[1][1], y, linewidth=linewidth, linestyle='--', color='r')


def coordinate2inx(coordinate, row=8, col=16, im_shape=[300, 600]):
    """Convert coordinate of coordinate of top-left corner into index. Index order is:
    [[0 1 2]
     [3 4 5]
     [6 7 8]]

    row, col: int
    Number of rows/columns of solar module

    im_shape: list
    [height, width] of perspective transformed solar module

    Returns
    -------
    inx: int
    Index
    """
    inx = col * round(coordinate[1] / (im_shape[0] / row)) + round(coordinate[0] / (im_shape[1] / col))

    return inx


def classify_cells(ann_path, defects_inx_dic, row_col=[8, 16], shape=[300, 600]):
    """Classify solar cells based on the class of the annotation of bounding box on the solar module

    Parameters
    ----------
    ann_path: str or pathlib.PosixPath
    Path of annotation file

    defects_inx_dic: dict
    Dict of defects with empty value. Used to store the index of the defective cells
    e.g.
    defects_dic = {
        'crack_bbox': [],
        'solder_bbox': [],
        'intra_bbox': [],
        'oxygen_bbox': []
    }

    row_col: list
    [row, col] of solar module

    shape: list
    [height, width] of solar module image

    Returns
    -------
    defects_inx_dic:
    Dict of defects with the index of the defective cells
    """
    with open(ann_path, 'r') as file:
        data = json.load(file)
    if len(data['objects']) > 0:
        for obj in data['objects']:
            classTitle = obj['classTitle']
            points = obj['points']['exterior'][0]
            inx = coordinate2inx(points, row=row_col[0], col=row_col[1], im_shape=shape)
            defects_inx_dic[classTitle].append(inx)

    return defects_inx_dic


def write_cells(single_cells, defects_inx, defect2folder, name, save_path):
    """Output cells into folders of corresponding class

    Parameters
    ----------
    single_cells: array
    Images of cropped cells

    defects_inx: dict
    Dict of defects with the index of the defective cells

    defect2folder: dict
    Convert keys in defects_inx into names of folders
    e.g.
    {
        "crack_bbox": "crack",
        "oxygen_bbox": "oxygen",
        "solder_bbox": "solder",
        "intra_bbox": "intra"
    }

    name: str
    Name of the solar module

    save_path: str or pathlib.PosixPath
    Path to store the output
    """
    for inx, cell in enumerate(single_cells):
        flag = True
        for key, value in defects_inx.items():
            if inx in value:
                cv.imwrite(str(Path(save_path) / f"{defect2folder[key]}/{name}__{str(inx)}.png"), cell)
                flag = False
        """
        if inx in defects_inx['crack_bbox']:
            cv.imwrite(path + 'crack/' + name + '__' + str(inx) + '.png', cell)
            flag = False
        if inx in defects_inx['solder_bbox']:
            cv.imwrite(path + 'solder/' + name + '__' + str(inx) + '.png', cell)
            flag = False
        if inx in defects_inx['oxygen_bbox']:
            cv.imwrite(path + 'oxygen/' + name + '__' + str(inx) + '.png', cell)
            flag = False
        if inx in defects_inx['intra_bbox']:
            cv.imwrite(path + 'intra/' + name + '__' + str(inx) + '.png', cell)
            flag = False
        """
        if flag:
            cv.imwrite(str(Path(save_path) / f"intact/{name}__{str(inx)}.png"), cell)



import json


def count_defects(ann_path, defect_dict):
    """count the number of defects on solar module

    Parameters
    ----------
    ann_path: str or pathlib.PosixPath
    The path of the json files that contain the information of predictions.
    Format should follow supervisely prediction format.

    defect_dict: dict
    The defects to be detected. Key is the defect name in the prediction file, and value is the number of defects.
    E.g.
    {'crack_bbox_yolo': 0, 'intra_bbox_yolo': 0, 'oxygen_bbox_yolo': 0, 'solder_bbox_yolo': 0}

    Returns
    -------
    defect_dict: dict
    The counted number of defects
    """
    with open(ann_path, 'r') as f:
            data = json.load(f)

    for defect in data["objects"]:
        if defect["classTitle"] in defect_dict.keys():
            defect_dict[defect["classTitle"]] += 1

    return defect_dict



from pathlib import Path
from imutils import paths
from tqdm import tqdm
import os
import shutil


def subfolder_merge(parent_folder, output_folder):
    """Merge the subfolders into one folder.

    Parameters
    ----------
    parent_folder: str or pathlib.PosixPath
    The folder that contains the subfolders

    output_folder: str or pathlib.PosixPath
    The folder that collects files from subfolders
    """
    os.makedirs(output_folder, exist_ok=True)
    file_paths = list(paths.list_files(str(parent_folder)))

    for file in tqdm(file_paths):
        shutil.copy(file, output_folder)
    
    os.remove(Path(output_folder)/'.DS_Store')



import numpy as np
import cv2 as cv

from sklearn.ensemble import RandomForestClassifier
from tqdm import tqdm

from sklearn import utils as sk_utils
from sklearn.metrics import f1_score

import itertools

import copy



def im_flatten(images):
    """Flatten the grayscale image

    Parameters
    ----------
    images: array
    The array of grayscale images

    Returns
    -------
    flattened_images: array
    Matrix, row of which is a flattened image
    """
    return images.reshape(len(images), -1)


def im_aug(images, labels, shuffle=True, select_aug=False, select_inx=None):
    """Do data augmentation, including vertical flip, horizontal flip and 180 rotation

    Parameters
    ----------
    images: array
    Array of grayscale images

    labels: list or array
    Labels of the images

    shuffle: bool
    If true, the augmented images will be shuffled

    select_aug: bool
    If true, only the images with labels in select_inx will be argumented

    select_inx: list
    Labels of the images to be augmented

    Returns
    -------
    x: array
    Array of augmented images

    y: array
    Array of labels
    """
    v_flip = []
    h_flip = []
    rotate = []
    labels_few = []

    if not select_aug:
        for image in images:
            v_flip.append(cv.flip(image, 0))
            h_flip.append(cv.flip(image, 1))
            rotate.append(cv.rotate(image, cv.ROTATE_180))
    elif select_aug:
        for i, image in enumerate(images):
            if labels[i] in select_inx:
                v_flip.append(cv.flip(image, 0))
                h_flip.append(cv.flip(image, 1))
                rotate.append(cv.rotate(image, cv.ROTATE_180))
                labels_few.append(labels[i])

    v_flip = np.array(v_flip)
    h_flip = np.array(h_flip)
    rotate = np.array(rotate)
    if not select_aug:
        x, y = np.concatenate((images, v_flip, h_flip, rotate)), np.concatenate((labels, labels, labels, labels))
    elif select_aug:
        labels_few = np.array(labels_few)
        x, y = np.concatenate((images, v_flip, h_flip, rotate)), np.concatenate(
            (labels, labels_few, labels_few, labels_few))

    if shuffle:
        return sk_utils.shuffle(x, y)
    else:
        return x, y


def iterate_values(S):
    """An iteration tool to return combined hyperparameters

    Parameters
    ----------
    S: dict
    The dictionary of hyperparameters

    Returns
    -------
    Combined hyperparameters: generator
    The generator of combined hyperparameters dict
    """
    keys, values = zip(*S.items())

    for row in itertools.product(*values):
        yield dict(zip(keys, row))


def train_val_rf(X_train, y_train, X_val, y_val, rf_para):
    """Train and validate random forest. Use macro F1 score to validate

    Parameters
    ----------
    X_train, X_val: array
    Matrix of training X, validation X.

    y_train, y_val: array
    Array of training y, validation y

    rf_para: dict
    Dictionary of random forest parameters

    Returns
    -------
    rf: RandomForestClassifier class
    Trained model

    f1_macro: float
    Macro F1 score on validation set
    """
    rf = RandomForestClassifier(**rf_para, n_jobs=-1)
    rf.fit(X_train, y_train)

    pred_val = rf.predict(X_val)
    f1_macro = f1_score(y_val, pred_val, average='macro')

    return rf, f1_macro


def random_search_rf(images_train, y_train, images_val, y_val, iteration, para_grid):
    """Hyperparameter tuning using random search

    Parameters
    ----------
    images_train, images_val: array
    Array of training images, validation images

    y_train, y_val: array
    Array of training labels, validation labels

    iteration: int
    Iteration of hyperparameter tuning, i.e., number of combination of hyperparameters

    para_grid: dict
    Dictionary of hyperparameters
    e.g. rf_para_grid = {
            'n_estimators': [10, 20, 30, 40, 50, 80, 100, 200, 400, 1000],
            'criterion': ['gini', 'entropy'],
            'max_depth': [10, 20, 40, 60, 80, 100, None],
            'bootstrap': [True, False]
        }

    Returns
    -------
    best_model: RandomForestClassifier class
    Trained model of optimal macro F1 score on validation set

    best_rf_para: dict
    Dictionary of parameters of model of optimal macro F1 score on validation set

    best_score: float
    Optimal macro F1 score on validation set
    """
    X_train = im_flatten(images_train)
    X_val = im_flatten(images_val)

    rf_para_list = list(iterate_values(para_grid))
    np.random.seed(42)
    rf_para_selected = np.random.choice(rf_para_list, iteration, replace=False)

    best_rf_para = {}
    best_score = 0
    for rf_para in tqdm(rf_para_selected):
        model, current_score = train_val_rf(X_train, y_train, X_val, y_val, rf_para)
        if current_score >= best_score:
            best_score = current_score
            best_rf_para = rf_para
            best_model = copy.deepcopy(model)
    return best_model, best_rf_para, best_score




import cv2 as cv


def im_rotate(im_path, rotate_angle):
    """ rotate input images through 90 or 180 degrees

    Parameters
    ----------
    im_path: str or pathlib.PosixPath
        The file path of the original image.

    rotate_angle: int, one of [0, 1, 2]
        The angle of rotation.
        0 = 90-clockwise,
        1 = 180
        2 = 90-counter-clockwise

    Returns
    -------
    image_r: array
    The rotated image.
    """
    image = cv.imread(str(im_path))
    image_r = cv.rotate(image, rotate_angle)

    return image_r



import numpy as np
import cv2 as cv
import json
import unittest
import matplotlib.pyplot as plt
import pv_vision.transform_crop.perspective_transform as pt
import pv_vision.transform_crop.cell_crop as crop
from pathlib import Path


class TestPerspectiveTransform(unittest.TestCase):

    img_only = Path("testing/img_only")
    img_mask = Path("testing/img_mask")

    def testImageThreshold(self):
        image = cv.imread(str(self.img_only/"M2001-0013_48432mV_9135mA_19.2Sec.jpg"), 0)
        img_thre_ad = pt.image_threshold(image, adaptive=True)
        img_thre = pt.image_threshold(image, threshold=50, adaptive=False)

        plt.figure(figsize=(10, 15))
        plt.subplot(311)
        plt.imshow(image, "gray")
        plt.subplot(312)
        plt.imshow(img_thre_ad, "gray")
        plt.subplot(313)
        plt.imshow(img_thre, "gray")
        plt.show()
        plt.clf()
        self.assertEqual(255, img_thre_ad.max())
        self.assertEqual(255, img_thre.max())

class TestCellCrop(unittest.TestCase):
    img_only = Path("testing/img_only")
    img_mask = Path("testing/img_mask")

    def testLineDetection(self):
        image = cv.imread(str(self.img_only / "PA3.jpg"), 0)
        image_thre = pt.image_threshold(image, adaptive=True)
        plt.figure(figsize=(10, 15))
        plt.subplot(211)
        plt.imshow(image, "gray")
        plt.subplot(212)
        plt.imshow(image_thre, "gray")
        plt.show()
        plt.clf()

        vline_abs = crop.detect_vertical_lines(image_thre, column=12)
        hline_abs = crop.detect_horizon_lines(image_thre, row=6, busbar=4)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import cv2 as cv
import pandas as pd


def load_colormap(map_path, gray_col, rgb_col, convert):
    """Output the colormap between grayscale and RGB

    Parameters
    ----------
    map_path: str or pathlib.PosixPath
    Path of colormap file

    gray_col: str
    Name of grayscale value column

    rgb_col: list
    List of names of RGB value column

    convert: int [0 or 1]
    If 0: rgb2gray
    If 1: gray2rgb

    Returns
    -------
    convert_dict: dict
    Mapping dictionary
    """
    colormap = pd.read_csv(map_path)
    grayscale = colormap[gray_col]
    rgb = colormap[rgb_col]

    convert_dict = {}
    if convert == 0:
        # rgb2gray
        for i in range(len(rgb)):
            convert_dict[str(list(rgb.iloc[i].to_numpy()))] = grayscale.iloc[i]
    elif convert == 1:
        # gray2rgb
        for i in range(len(grayscale)):
            convert_dict[grayscale.iloc[i]] = rgb.iloc[i].to_numpy()

    return convert_dict


def convert_rgb2gray(image, convert_dic):
    """convert rgb image to grayscale

    Parameters
    ----------
    image: array
    RGB image. Channel order should be RGB.

    convert_dic: dict
    dictionary key is str(rgb list), value is grayscale value

    Returns
    -------
    image_gray: array
    Grayscale image
    """
    image_r = image[:, :, 0]
    image_g = image[:, :, 1]
    image_b = image[:, :, 2]
    im_shape = image_r.shape

    image_gray = np.zeros(im_shape)

    for i in range(im_shape[0]):
        for j in range(im_shape[1]):
            image_gray[i, j] = convert_dic[str([image_r[i, j], image_g[i, j], image_b[i, j]])]

    return image_gray


def convert_gray2rgb(image, convert_dic):
    """convert rgb image to grayscale

        Parameters
        ----------
        image: array
        Grayscale image. Only 1 channel.

        convert_dic: dict
        dictionary key is int(grayscale value), value is rgb array

        Returns
        -------
        image_rgb: array
        RGB image
        """
    im_shape_0 = image.shape[0]
    im_shape_1 = image.shape[1]
    image_rgb = np.zeros((im_shape_0, im_shape_1, 3))

    for i in range(im_shape_0):
        for j in range(im_shape_1):
            image_rgb[i, j, :] = convert_dic[image[i, j]]

    return image_rgb


def remove_repeated_channel(image):
    """Remove repeated channel in grayscale images.
       Opencv read grayscale images into 3-channel if not specified, so sometimes people may output a 3-channel
       grayscale image.

    Parameters
    ----------
    image: array
    Grayscale images with 3 repeated channels

    Returns
    -------
    image: array
    Grayscale images with 1 channel
    """
    return image[:, :, 0]




import numpy as np
import matplotlib.pyplot as plt
import cv2 as cv

from sklearn.metrics import f1_score
from tqdm import tqdm
from PIL import Image
from imutils.paths import list_images
import torch
import torch.nn as nn
from torchvision import transforms
from torch.utils.data import Dataset
import copy, random

import os


class OneRotationTransform:
    """Rotate by one given angles."""

    def __init__(self, angles):
        self.angles = angles

    def __call__(self, x):
        angle = random.choice(self.angles)
        return transforms.functional.rotate(x, angle)


class SolarDataset(Dataset):
    def __init__(self, images, labels, transform, transform2=None, inx_aug=None):
        self.images = images
        self.labels = torch.from_numpy(labels)
        self.transform = transform
        self.transform2 = transform2
        self.inx_aug = inx_aug

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        image = Image.fromarray(cv.cvtColor(self.images[idx], cv.COLOR_BGR2RGB))
        label = self.labels[idx]
        if self.transform2:
            if label in self.inx_aug:
                image = self.transform(image)
            else:
                image = self.transform2(image)
        else:
            image = self.transform(image)

        return image, label


class PredDataset(Dataset):
    # Only used for prediction
    # No labels
    def __init__(self, images, transform):
        self.images = images
        self.transform = transform

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        image = Image.fromarray(cv.cvtColor(self.images[idx], cv.COLOR_BGR2RGB))
        image = self.transform(image)

        return image


def load_data(im_dir, store_name=False):
    """load the images and labels

    Parameters
    ----------
    im_dir: str
    Folder name of images. Folder should be like: 'train/category_i/image.png' and use 'train' as im_dir

    store_name: bool
    If True, store the name of solar modules

    Returns
    -------
    images, labels, names: list
    List of images, labels and names if store_name==True
    """
    images = []
    labels = []
    if store_name:
        names = []

        for im_path in tqdm(list(list_images(im_dir))):
            images.append(cv.imread(im_path))
            labels.append(im_path.split('/')[-2])
            names.append(os.path.splitext(os.path.split(im_path)[-1])[0])

        return images, labels, names
    else:
        for im_path in tqdm(list(list_images(im_dir))):
            images.append(cv.imread(im_path))
            labels.append(im_path.split('/')[-2])

        return images, labels


def train_model(model, trainloader, valloader, solar_train, solar_val, criterion, optimizer, scheduler, device, num_epochs=25):
    """Train a neural network using pytorch. Different loss functions are compared

    Parameters
    ----------
    model:
    Pytorch model

    trainloader, valloader: 
    Pytorch dataloader

    solar_train, solar_val:
    Pytorch dataset

    criterion:
    Loss function

    optimizer:
    optimizer

    scheduler:
    Learning rate scheduler

    num_epochs: int
    Number of epochs

    device: torch.device()

    Returns
    -------
    weights_dict: dict
    The model weights on optimal epochs. Selection can be based on optimal accuracy, loss or f1 score of validation set

    loss_acc: dict
    Record of loss, accuracy and f1 score during training and validation process
    """
    loss_acc = {
        'train': {"loss": [], "acc": [], "f1_macro": []},
        'val': {"loss": [], "acc": [], "f1_macro": []}
    }

    best_acc_model_wts = copy.deepcopy(model.state_dict())
    best_acc = 0.0
    best_loss = 1e6
    best_f1 = 0.0

    for epoch in range(num_epochs):
        print('Epoch {}/{}'.format(epoch, num_epochs - 1))
        print('-' * 10)

        # Each epoch has a training and validation phase
        for phase in ['train', 'val']:
            if phase == 'train':
                model.train()  # Set model to training mode
                dataloader = trainloader
                dataset_size = len(solar_train)
            else:
                model.eval()  # Set model to evaluate mode
                dataloader = valloader
                dataset_size = len(solar_val)

            running_loss = 0.0
            running_corrects = 0

            # Iterate over data.
            if phase == 'val':  # collect all the labels in one epoch
                preds_val = []
                labels_val = []
            elif phase == 'train':
                preds_train = []
                labels_train = []
                # prob_val = []
                # m = nn.Softmax(dim=1)
            for inputs, labels in tqdm(dataloader):
                # collect val label to compute f1
                if phase == 'val':
                    labels_val = np.append(labels_val, labels.numpy())
                elif phase == 'train':
                    labels_train = np.append(labels_train, labels.numpy())
                inputs = inputs.to(device)
                labels = labels.long()
                labels = labels.to(device)

                # zero the parameter gradients
                optimizer.zero_grad()

                # forward
                # track history if only in train
                with torch.set_grad_enabled(phase == 'train'):
                    outputs = model(inputs)
                    _, preds = torch.max(outputs, 1)
                    loss = criterion(outputs, labels)

                    # backward + optimize only if in training phase
                    if phase == 'train':
                        preds_train = np.append(preds_train, preds.cpu().numpy())
                        loss.backward()
                        optimizer.step()

                    # record val prediction to compute f1
                    if phase == "val":
                        # prob_val= np.append(prob_val, m(outputs.cpu()).numpy())
                        preds_val = np.append(preds_val, preds.cpu().numpy())

                # statistics
                running_loss += loss.item() * inputs.size(0)
                running_corrects += torch.sum(preds == labels.data)

            if phase == 'train':
                scheduler.step()

            epoch_loss = running_loss / dataset_size
            epoch_acc = running_corrects.double() / dataset_size
            loss_acc[phase]['loss'].append(epoch_loss)
            loss_acc[phase]['acc'].append(float(epoch_acc))

            print('{} Loss: {:.4f} Acc: {:.4f}'.format(
                phase, epoch_loss, epoch_acc))

            # compute f1
            if phase == 'val':
                # labels_val_binary = label_binarize(labels_val, classes = le.transform(le.classes_))
                # prob_val = np.array(prob_val)
                # epoch_prc_micro = average_precision_score(labels_val_binary.ravel(), prob_val.ravel(), average='micro')
                epoch_f1 = f1_score(labels_val, preds_val, average='macro')
                loss_acc[phase]['f1_macro'].append(epoch_f1)
                print('F1: {:4f}'.format(epoch_f1))
            elif phase == 'train':
                epoch_f1 = f1_score(labels_train, preds_train, average='macro')
                loss_acc[phase]['f1_macro'].append(epoch_f1)
                print('F1: {:4f}'.format(epoch_f1))

            # deep copy the model
            if phase == 'val' and epoch_acc > best_acc:
                best_acc = epoch_acc
                best_acc_model_wts = copy.deepcopy(model.state_dict())
            if phase == 'val' and epoch_loss < best_loss:
                best_loss = epoch_loss
                best_loss_model_wts = copy.deepcopy(model.state_dict())
            if phase == 'val' and epoch_f1 > best_f1:
                best_f1 = epoch_f1
                best_f1_model_wts = copy.deepcopy(model.state_dict())

        print()

    print('Best val Acc: {:4f}'.format(best_acc))
    print('Best val F1: {:4f}'.format(best_f1))

    return {'best_acc': best_acc_model_wts, 'best_loss': best_loss_model_wts, 'best_f1': best_f1_model_wts}, loss_acc


def training_plot(epochs, loss_accuracy, metric='loss'):
    """Plot the metrics change during training/val process

    Parameters
    ----------
    epochs: int
    Number of epochs

    loss_accuracy: dict
    Record of loss, accuracy and f1 score during training and validation process

    metric: str
    Metrics to show
    """
    plt.plot(range(epochs), loss_accuracy['train'][metric], label=f'train_{metric}')
    plt.plot(range(epochs), loss_accuracy['val'][metric], label=f'val_{metric}')
    plt.legend()


def predict_test(testloader, model_fit, device):
    """Make prediction on the testing set

    Parameters
    ----------
    testloader: torch.utils.data.dataloader.DataLoader
    Dataloader of testing set

    model_fit: Pytorch model
    Machine learning model used to make prediction

    device: torch.device
    Device to work on

    Returns
    -------
    pred_test: list
    Predicted labels

    prob_test: list
    Probability of each category
    """
    pred_test = []
    prob_test = []

    m = nn.Softmax(dim=1)
    with torch.no_grad():
        for input in tqdm(testloader):
            input = input.to(device)

            output = model_fit(input)
            prob_test += m(output.cpu()).tolist()
            _, pred = torch.max(output, 1)
            pred_test += pred.cpu().tolist()
    return pred_test, prob_test





import numpy as np


def extract_length(line):
    line_temp = line[1:]
    line_temp.append(line[-1])
    length = np.sqrt(np.sum((line_temp - np.array(line))**2, axis=1))
    length_tot = np.sum(length)

    return length_tot


def extract_angle(line):
    x = list(zip(*line))[0]
    y = list(zip(*line))[1]
    k, b = np.polyfit(x, y, 1)

    return k, b


def flatten_list(list_comb):
    list_flatten = []
    list_comb = list(list_comb)
    for element in list_comb:
        list_flatten += element

    return list_flatten



import csv
import os
import shutil
import argparse
from pathlib import Path
from tqdm import tqdm


def im_move(im_folder_path, csv_path, subfolders, parentfolder='classified_images', im_extension=None):
    """Move the source images into subfolders based on their categories.

    Parameters
    ----------
    im_folder_path: str or pathlib.PosixPath
        The folder path of the original images.

    csv_path: str or pathlib.PosixPath
        The path of the csv file that indicates the category of each solar module.
        The first column of the csv file should be the name of a module without filename extension
        The second column should be its category which is the same in the subfolders.

    subfolders: list of strings
        The categories of the solar modules.
        E.g. ['category1', 'category2', 'category3']

    parentfolder: str
        The parent folder name of subfolders

    im_extension: str
        The filename extension of the src images, e.g. '.png', '.jpg', etc.
    """
    if not im_extension:
        image = os.listdir(im_folder_path)[0]
        im_extension = os.path.splitext(image)[-1]

    folder = Path(parentfolder)
    for subfolder in subfolders:
        os.makedirs(folder/ subfolder, exist_ok=True)

    with open(csv_path, 'r') as file:
        data = [line.rstrip() for line in file]

    for cell in tqdm(data):
        name, label = cell.split(',')[0], cell.split(',')[1]
        shutil.copy(im_folder_path/ (name + im_extension), folder/label)



import cv2 as cv
import json


def draw_bbox(ann_path, image, color, thickness=3):
    """Draw bounding box of defective cells on the solar module images

    Parameters
    ----------
    ann_path: str or pathlib.PosixPath
    The path of the prediction json files from supervisely

    image: array
    The image of a solar module that is perspectively transformed

    color: dict
    The color of bbox.
    E.g.
    {"crack_bbox_yolo": [60, 124, 90],
     "solder_bbox_yolo": [190, 112, 78],
     "oxygen_bbox_yolo": [40, 64, 183],
     "intra_bbox_yolo": [103, 52, 154]}

     thickness: int
     The thickness of the bbox. Default is 3

     Returns
     -------
     function directly operates on the input image
    """
    with open(ann_path, 'r') as file:
        data = json.load(file)
    if len(data['objects']) > 0:
        for obj in data['objects']:
            classTitle = obj['classTitle']
            points = obj['points']['exterior']

            cv.rectangle(image, tuple(points[0]), tuple(points[1]), color[classTitle], thickness)


from setuptools import setup, find_packages
from pv_vision import __version__

with open("long_description.md", "r") as f:
    long_description = f.read()

setup(
    name="pv_vision",
    version=__version__,
    author="XinChen",
    author_email="chenxin0210@lbl.gov",
    description="Image analysis of defects on solar modules, including automatic detection and power loss prediction",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hackingmaterials/pv-vision",
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.8',
    install_requires=[
        "matplotlib>=3.4.2",
        "opencv_python>=4.4.0.44",
        "seaborn>=0.11.2",
        "torchvision>=0.7.0",
        "pandas>=1.3.2",
        "imutils>=0.5.3",
        "scipy>=1.6.0",
        "numpy>=1.19.2",
        "tqdm>=4.56.0",
        "torch>=1.6.0",
        "Pillow>=9.0.1",
        "scikit_learn>=0.24.2",
    ],
    license="BSD",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
)



import numpy as np
import pv_vision.transform_crop.perspective_transform as transform
from skimage.morphology import skeletonize
from pathlib import Path
import os
import cv2 as cv
import json
from scipy import stats


def skeleton_crack(mask_crack):
    """Skeletonize crack masks
    Parameteres
    -----------
    mask_crack: array
    Mask of cracks

    Returns
    -------
    ske_crack: array
    Skeletonized crack mask
    """
    return skeletonize(mask_crack).astype(np.uint8)


def extend_busbar(mask_busbar, kernel_size=(10, 100)):
    """Connet and extend broken busbars. Return the skeleton of the busbar masks
    Parameters
    ----------
    mask_busbar: array
    Masks of busbars

    kernel_size: list or tuple
    Kernel used to do morphological operation. Details available on
    https://docs.opencv.org/4.5.4/d9/d61/tutorial_py_morphological_ops.html

    Returns
    -------
    ske_busbar: array
    Skeletonized busbar masks
    """
    kernel = np.ones(kernel_size, np.uint8)
    closing = cv.morphologyEx(mask_busbar, cv.MORPH_CLOSE, kernel)
    ske_busbar = skeletonize(closing).astype(np.uint8)

    return ske_busbar


def locate_busbar(ske_busbar):
    """Get position of busbars
    Parameters
    ----------
    ske_busbar: array
    Skeletonized busbar masks

    Returns
    -------
    pos_busbar: list
    Positions of busbars
    """
    numlist_busbar = []
    for i in np.linspace(10, ske_busbar.shape[-1] - 10, 10, dtype=int):
        numlist_busbar.append(len(np.argwhere(ske_busbar[:, i] == 1)))

    num_busbar = stats.mode(numlist_busbar).mode[0]

    pos_busbar = np.zeros((num_busbar, 1))
    for i in np.linspace(10, ske_busbar.shape[-1] - 10, 100, dtype=int):
        tem_pos = np.argwhere(ske_busbar[:, i] == 1)

        if len(tem_pos) == num_busbar:
            pos_busbar = np.hstack((pos_busbar, tem_pos))

    pos_busbar = np.delete(pos_busbar, 0, axis=1)
    pos_busbar = pos_busbar.mean(axis=1, dtype=int).tolist()

    return pos_busbar


def skeleton_cell(ske_crack, pos_busbar):
    """Get the skeleton of a cell. Crack has the value of -1, busbar is 1, other area is 0.
    Parameters
    ----------
    ske_crack: array
    Skeleton of crack masks

    pos_busbar: list
    Position of busbars

    Returns
    -------
    ske_cell: array
    Skeleton of cell
    """
    ske_cell = ske_crack * -1
    for i in pos_busbar:
        ske_cell[i, :] = 1

    return ske_cell


def stop_diff(val):
    """Check whether diffusion should stop. If meet busbar(val=1) or crack(val=-1), stop diffusion
    Parameters
    ----------
    val: int
    Grayscale value of a pixel

    Returns
    -------
    bool
    """
    return val == 1 or val == -1


def diff_up(image, row, col):
    """Diffuse the electrons up. The busbars are horizontally aligned
    Parameters
    ----------
    image: array
    Skeleton of cell

    row, col: int
    Position of current pixel
    """
    current = row - 1
    while not (current < 0 or stop_diff(image[current, col])):
        image[current, col] = 1
        current -= 1


def diff_down(image, row, col):
    """Diffuse the electrons down. The busbars are horizontally aligned
    Parameters
    ----------
    image: array
    Skeleton of cell

    row, col: int
    Position of current pixel
    """
    end = image.shape[0]
    current = row + 1
    while not (current > end - 1 or stop_diff(image[current, col])):
        image[current, col] = 1
        current += 1


def diffuse_line(image, row):
    """Diffuse the electrons from one busbar. The busbars are horizontally aligned
    Parameters
    ----------
    image: array
    Skeleton of cell

    row: int
    Position of current busbar
    """
    for j in range(image.shape[-1]):
        diff_up(image, row, j)
        diff_down(image, row, j)


def diffuse(image, pos_busbar):
    """Diffuse the electrons from all busbars. The busbars are horizontally aligned
    Parameters
    ----------
    image: array
    Skeleton of cell

    pos_busbar: list
    Position of busbars

    Returns
    -------
    image_c: array
    Diffused cell
    """
    image_c = np.copy(image)
    for i in pos_busbar:
        diffuse_line(image_c, i)
    return image_c


def count_area(cell_diff):
    """Count worst-case percentage of inactive area
    Parameters
    ----------
    cell_diff: array
    Diffused cell image

    Returns
    -------
    percentage of inactive area: float
    """
    inactive_area = np.zeros(cell_diff.shape).astype(np.uint8)
    inactive_area[cell_diff == 0] = 1
    return inactive_area.sum() / (inactive_area.shape[0] * inactive_area.shape[1])


def detect_inactive(mask_crack, mask_busbar, extend_kernel=(10, 100)):
    """Detect the worst-case isolated area and calculate its proportion
    Parameters
    ----------
    mask_crack: array
    Mask of cracks

    mask_busbar: array
    Masks of busbars

    extend_kernel: list or tuple
    Kernel used to do morphological operation to extend busbar. Details available on
    https://docs.opencv.org/4.5.4/d9/d61/tutorial_py_morphological_ops.html

    Returns
    -------
    inactive_area: array
    Binary isolated area

    inactive_prop: float
    percentage of inactive area. Not in the form of %
    """
    ske_crack = skeleton_crack(mask_crack)
    ske_busbar = extend_busbar(mask_busbar, kernel_size=extend_kernel)
    pos_busbar = locate_busbar(ske_busbar)
    ske_cell = skeleton_cell(ske_crack, pos_busbar)
    cell_diff = diffuse(ske_cell, pos_busbar)

    inactive_area = np.zeros(cell_diff.shape).astype(np.uint8)
    inactive_area[cell_diff == 0] = 1
    inactive_prop = inactive_area.sum() / (inactive_area.shape[0] * inactive_area.shape[1])

    return inactive_area, inactive_prop




import pv_vision.transform_crop.perspective_transform as transform
import pv_vision.transform_crop.cell_crop as seg
import numpy as np
import cv2 as cv
from pathlib import Path


class AbstractModule:
    """Parent class. This class provide basic methods of processing a module image"""

    def __init__(self, image, row, col, busbar):
        self._image = image
        self._size = image.shape
        self._row = row
        self._col = col
        self._busbar = busbar

    @property
    def image(self):
        """The EL image of the module"""
        return self._image

    @image.setter
    def image(self, new_image):
        self._image = new_image
        self._size = new_image.shape

    @property
    def size(self):
        """The size of the image"""
        return self._size

    @property
    def row(self):
        """The number of row and col of the module"""
        return self._row

    @row.setter
    def row(self, row):
        self._row = row

    @property
    def col(self):
        """The number of row and col of the module"""
        return self._col

    @col.setter
    def col(self, col):
        self._col = col

    @property
    def busbar(self):
        """The number of busbars"""
        return self._busbar

    @busbar.setter
    def busbar(self, busbar):
        self._busbar = busbar

    def _reset(self, new_image, new_row, new_col):
        """Reset the image in this instance"""
        self._image = new_image
        self._size = new_image.shape
        self._row = new_row
        self._col = new_col

    def resize(self, new_size, in_place=False):
        """Resize the image

        Parameters
        ----------
        new_size: list or tuple
        [height, width] of the resized image

        in_place: bool
        If true, change will be applied to current instance

        Returns
        -------
        image_r: array
        Resized image
        """
        image_r = cv.resize(self._image, (new_size[1], new_size[0]))
        if in_place:
            self._reset(image_r, self._row, self._col)
        else:
            return image_r

    def rotate(self, rotate_angle, in_place=False):
        """ rotate input images through 90 or 180 degrees

        Parameters
        ----------
        rotate_angle: int, one of [0, 1, 2]
            The angle of rotation.
            0 = 90-clockwise,
            1 = 180
            2 = 90-counter-clockwise

        in_place: bool
        If true, change will be applied to current instance

        Returns
        -------
        image_r: array
        The rotated image.
        """
        image_r = cv.rotate(self._image, rotate_angle)
        if in_place:
            if rotate_angle == 1:
                self._reset(image_r, self._row, self._col)
            else:
                self._reset(image_r, self._col, self._row)
        else:
            return image_r

    def remove_channel(self, in_place=False):
        """Remove repeated channel in grayscale images.
           Opencv read grayscale images into 3-channel if not specified, so sometimes people may output a 3-channel
           grayscale image.

        Parameters
        ----------
        in_place: bool
        If true, change will be applied to current instance

        Returns
        -------
        image: array
        Grayscale images with 1 channel
        """
        if len(self._size) == 2:
            print("It is already 1-channel")
        else:
            if in_place:
                self._reset(self._image[:, :, 0], self._row, self._col)
            else:
                return self._image[:, :, 0]

    def copy_channel(self, in_place=False):
        """Duplicate the grayscale channel and expand to 3 channels

        Parameters
        ----------
        in_place: bool
        If true, change will be applied to current instance

        Returns
        -------
        image: array
        Grayscale images with 3 channels
        """
        if len(self._size) == 3:
            print("It is already 3-channel")
        else:
            image_r = cv.merge((self._image, self._image, self._image))
            if in_place:
                self._reset(image_r, self._row, self._col)
            else:
                return image_r

    def save_fig(self, save_path):
        """Save the image file

        Parameters
        ----------
        save_path: str or pathlib.PosixPath
        The folder path of the original images.
        """
        cv.imwrite(str(save_path), self._image)


class SplitModule(AbstractModule):
    """Raw module image class that use splitting to crop cells"""

    def __init__(self, image, row, col, busbar):
        super().__init__(image, row, col, busbar)
        self._cells = None
        self._vline_abs = None
        self._hline_abs = None

    @property
    def cells(self):
        if self._cells is None:
            print("Cells are not cropped")

        return self._cells

    def crop_cell(self, cellsize, hsplit=100, hthre=0.8, hinterval=100, hmargin=None, vsplit=50, vthre=0.6,
                  vinterval=250, vmargin=None, savepath=None, displace=None):
        """Crop cells

        Parameters
        ----------
        cellsize: int
        Edge size of a cell. This can be an estimated value from raw module image.

        hsplit, vplit: int
        Number of horizontal/vertical splits

        hthre, vthre: float
        Peaks will be set as 1 above this threshold

        savepath: str
        Save path

        displace: int
        Displace the detected lines

        Returns
        -------
        cells: array
        """

        image_thre = transform.image_threshold(self.image, adaptive=True)
        image_thre = cv.resize(image_thre, (4000, 2500))
        self._vline_abs = seg.detect_vertical_lines(image_thre, cell_size=cellsize,
                                                    column=self.col, thre=hthre, split=hsplit, peak_interval=vinterval,
                                                    margin=vmargin)
        self._hline_abs = seg.detect_horizon_lines(image_thre, row=self.row, busbar=self.busbar,
                                                   cell_size=cellsize, thre=vthre, split=vsplit,
                                                   peak_interval=hinterval, margin=hmargin)

        self._cells = seg.segment_cell(self.image, self._hline_abs, self._vline_abs, cellsize, savepath, displace)

        return self._cells

    def plot_edges(self, linewidth=1):
        seg.plot_edges(self.image, self._hline_abs, self._vline_abs, linewidth)


class MaskModule(AbstractModule):
    """Raw module image class that has a mask of target modules"""

    def __init__(self, image, row, col, busbar):
        super().__init__(image, row, col, busbar)
        self._mask = None
        self._mask_center = None
        self._corners = None
        self._transformed = None
        self._cells = None

    @property
    def mask(self):
        if self._mask is None:
            print("Mask is not loaded")

        return self._mask

    @property
    def mask_center(self):
        if self._mask_center is None:
            print("Mask is not loaded")

        return self._mask_center

    @property
    def corners(self):
        if self._corners is None:
            print("Corners are not detected")

        return self._corners

    @property
    def transformed_image(self):
        if self._transformed is None:
            print("Module is not transformed")

        return self._transformed

    @property
    def cells(self):
        if self._cells is None:
            print("Cells are not cropped")

        return self._cells

    def load_mask(self, mask_path=None, output=False, blur=5, thre=None):
        """Load the mask of a module.
        If the background of the image looks good, you don't need to provide masks explicitly.
        Otherwise, you should provide masks predicted from neural networks

        Parameters
        ----------
        mask_path: str
        If provided, then mask is loaded explicitly. Otherwise, this method will detect masks automatically.

        output: bool
        If true, return mask and mask center

        blur: int
        Blur filter size in cv.medianBlur

        thre: int or float
        Threshold value in cv.threshold.
        Pixels above this value will be set 255, and below will be 0.
        If threshold < 1, first threshold (%) of the image grayscale value will be used.
        If not given, first 10% of the image grayscale value will be used.
        """
        if mask_path is None:
            self._mask = transform.image_threshold(image=self.image, blur=blur, threshold=thre)
        else:
            self._mask, self._mask_center = \
                transform.load_mask(mask_path, self._image, 'module_unet')

        if output:
            return self._mask

    def corner_detection_line(self, dist=200, displace=0, method=0,
                              corner_center=False, center_displace=10, output=False):
        """Detect the corner of solar module. Intersection of edges or corner_detection method is used.

        Parameters
        ----------
        mask: array
        Image of mask

        mask_center: array
        Coordinate of the mask center

        dist: int
        Distance threshold between two corners. Default is 200.

        displace: int
        Displacement of the detected corners to increase tolerance.

        method: int
        0 = use cv.goodFeaturesToTrack() to detect corners
        1 = use cv.HoughLines() to detect edges first and then find the intersections

        corner_center: Bool
        If True, use auto-detected nask center. Otherwise use 'mask_center' parameter. Default is False.

        center_displace: int
        Displacement of the mask center when dividing the mask into four corner parts

        output: bool
        If true, return the corners

        Returns
        -------
        Corners: array
        Sorted coordinates of module corners. The order is top-left, top-right, bottom-left and bottom-right
        """
        if self._mask is None:
            print("Mask is not loaded")
            return None
        self._corners = transform.find_module_corner(self._mask, self._mask_center,
                                                     dist, displace, method, corner_center, center_displace)
        if output:
            return self._corners

    def corner_detection_cont(self, mode=0, output=False):
        """ Detect the corners of a solar module

        Parameters
        ----------
        mode: int
        mode == 0: detect corners of the convex of module
        mode == 1: detect corners of the approximated convex of module
        mode == 2: detect corners of the approximated contour of the module
        mode == 3: detect corners of the blurred mask of the module
        mode == 4: detect corners using boudingRect. This is useful if your module is a rounded rectangle.

        output: bool
        If true, return the corners

        Returns
        -------
        Corners: array
        Corners of the solar module
        """
        if self._mask is None:
            print("Mask is not loaded")
            return None
        self._corners = transform.find_module_corner2(self._mask, mode)
        if output:
            return self._corners

    def transform(self, width=None, height=None, cellsize=None, auto_rotate=True, img_only=True):
        """Do perspective transform on the solar module

        Parameters
        ----------
        width, height: int
        Width/height of transformed image

        cellsize: int
        Edge length of a cell

        auto_rotate: bool
        If true, automatically adjust module orientation such that shorter side is vertically aligned.

        img_only: bool
        If true, only return the image of transformed module.
        Otherwise return a transformed module instance

        Returns
        -------
        wrap: array or instance
        Transformed solar module
        """
        if self._corners is None:
            print("Corners are not detected. Start automatic detection")
            return None
            # self._corners = self.corner_detection_cont(mode=1)

        if cellsize:
            width = self.col * cellsize
            height = self.row * cellsize
        wrap = transform.perspective_transform(self._image, self._corners, width, height, rotate=auto_rotate)
        self._transformed = TransformedModule(wrap, self._row, self._col, self._busbar)
        if img_only:
            return wrap
        else:
            return TransformedModule(wrap, self._row, self._col, self._busbar)

    def is_transformed(self, x_min, y_min):
        """Determine whether the module is properly transformed by checking the number of internal edges.

        Parameters
        ----------
        image: array
        The image that needs to be checked

        x_min, y_min: int
        The threshold of the number of detected internal edges. X means col and y means row

        Returns
        -------
        bool
        """
        res = self._transformed.is_transformed(x_min, y_min)
        return res

    def plot_peaks(self, n_list, split_size, interval, thre, margin, direction=0):
        self._transformed.plot_peaks(n_list, split_size, interval, thre, margin, direction)

    def crop_cell(self, cellsize, inner_edges_para=None, plot=False, simple=False):
        if inner_edges_para is not None:
            for para in ["vl_interval", "vl_thre", "vl_split_size", "vl_margin",
                            "hl_interval", "hl_thre", "hl_split_size", "hl_margin"]:
                if para not in inner_edges_para:
                    raise KeyError("{} is not provided.".format(para))
        cells = self._transformed.crop_cell(cellsize=cellsize,
                                            inner_edges_para=inner_edges_para,
                                            plot=plot, simple=simple)

        return cells


class TransformedModule(AbstractModule):
    """Solar module class after perspective transformation"""

    def __init__(self, image, row, col, busbar):
        super().__init__(image, row, col, busbar)
        if len(self._size) != 2:
            super().remove_channel(in_place=True)
        self.inner_edges_para = {}
        self.inner_edges_abs = {}

    # @staticmethod
    def is_transformed(self, x_min, y_min):
        """Determine whether the module is properly transformed by checking the number of internal edges.

        Parameters
        ----------
        image: array
        The image that needs to be checked

        x_min, y_min: int
        The threshold of the number of detected internal edges. X means col and y means row

        Returns
        -------
        bool
        """
        peak_x, peak_y = transform.find_inner_edge(self._image)
        return (len(peak_x) >= x_min) and (len(peak_y) >= y_min)

    def plot_peaks(self, n_list, split_size, interval, thre, margin, direction=0):
        seg.plot_peaks_three(n_list, image=self.image, busbar=self.busbar, split_size=split_size, direction=direction,
                             thre=thre, interval=interval, margin=margin)
        if direction==0:
            self.inner_edges_para['vl_interval'] = interval
            self.inner_edges_para['vl_thre'] = thre
            self.inner_edges_para['vl_split_size'] = split_size
            self.inner_edges_para['vl_margin'] = margin
        else:
            self.inner_edges_para['hl_interval'] = interval
            self.inner_edges_para['hl_thre'] = thre
            self.inner_edges_para['hl_split_size'] = split_size
            self.inner_edges_para['hl_margin'] = margin

    def _detect_inner_edges(self, cellsize, vl_interval, vl_thre, vl_split_size, vl_margin,
                            hl_interval, hl_thre, hl_split_size, hl_margin, simple=False):

        if simple:
            vline_abs = list(zip(np.zeros(self.col - 1),
                                 np.linspace(0, self.size[1], self.col + 1)[1: -1].astype(int)))
            hline_abs = list(zip(np.zeros(self.row - 1),
                                 np.linspace(0, self.size[0], self.row + 1)[1: -1].astype(int)))
        else:
            vinx_split, vline_split = seg.detect_edge(self._image, row_col=[self.row, self.col], cell_size=cellsize,
                                                      busbar=self.busbar, peaks_on=0, split_size=vl_split_size,
                                                      peak_interval=vl_interval, thre=vl_thre, margin=vl_margin)
            vline_abs = seg.linear_regression(vinx_split, vline_split)
            hinx_split, hline_split = seg.detect_edge(self._image, row_col=[self.row, self.col], cell_size=cellsize,
                                                      busbar=self.busbar, peaks_on=1, split_size=hl_split_size,
                                                      peak_interval=hl_interval, thre=hl_thre, margin=hl_margin)
            hline_abs = seg.linear_regression(hinx_split, hline_split)

        hline_abs_couple = seg.couple_edges(hline_abs, length=self.size[0])
        self.inner_edges_abs['hline_abs_couple'] = hline_abs_couple
        vline_abs_couple = seg.couple_edges(vline_abs, length=self.size[1])
        self.inner_edges_abs['vline_abs_couple'] = vline_abs_couple

        return hline_abs_couple, vline_abs_couple

    def plot_edges(self):
        if 'hline_abs_couple' not in self.inner_edges_abs or 'vline_abs_couple' not in self.inner_edges_abs:
            raise KeyError("Inner edge parameters are not detected.")
        hline_abs_couple = self.inner_edges_abs['hline_abs_couple']
        vline_abs_couple = self.inner_edges_abs['vline_abs_couple']
        seg.plot_edges(self.image, hline_abs_couple, vline_abs_couple, linewidth=1)

    def crop_cell(self, cellsize, inner_edges_para=None, plot=False, simple=False):
        if inner_edges_para is None:
            if 'vl_interval' not in self.inner_edges_para:
                raise KeyError("vertical edge parameter is not provided")
            elif 'hl_interval' not in self.inner_edges_para:
                raise KeyError("horizontal edge parameter is not provided")
            else:
                inner_edges_para = self.inner_edges_para

        hline_abs_couple, vline_abs_couple = self._detect_inner_edges(cellsize, simple=simple,
                                                                      **inner_edges_para)
        if plot:
            self.plot_edges()
        return np.array(seg.segment_cell(self.image, hline_abs_couple, vline_abs_couple, cellsize=cellsize))

    def classify_cells(self, ann_path, defects_inx_dic):
        """Classify solar cells based on the class of the annotation of bounding box on the solar module

        Parameters
        ----------
        ann_path: str or pathlib.PosixPath
        Path of annotation file

        defects_inx_dic: dict
        Dict of defects with empty value. Used to store the index of the defective cells
        e.g.
        defects_dic = {
            'crack_bbox': [],
            'solder_bbox': [],
            'intra_bbox': [],
            'oxygen_bbox': []
        }

        row_col: list
        [row, col] of solar module

        shape: list
        [height, width] of solar module image

        Returns
        -------
        defects_inx_dic:
        Dict of defects with the index of the defective cells
        """
        return seg.classify_cells(ann_path, defects_inx_dic, row_col=self.row_col, shape=self.size)

    @staticmethod
    def write_cells(single_cells, defects_inx, defect2folder, name, save_path):
        """Output cells into folders of corresponding class

        Parameters
        ----------
        single_cells: array
        Images of cropped cells

        defects_inx: dict
        Dict of defects with the index of the defective cells

        defect2folder: dict
        Convert keys in defects_inx into names of folders
        e.g.
        {
            "crack_bbox": "crack",
            "oxygen_bbox": "oxygen",
            "solder_bbox": "solder",
            "intra_bbox": "intra"
        }

        name: str
        Name of the solar module

        save_path: str or pathlib.PosixPath
        Path to store the output
        """
        seg.write_cells(single_cells, defects_inx, defect2folder, name, save_path)



import os.path

import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import fbeta_score
import pickle
from pathlib import Path
from sklearn import metrics
import seaborn as sns
import pandas as pd
import json


def predict_failed(y_test, pred_test, images_test, model_name=None, save_path=None):
    """Output failed predictions

    Parameters
    ----------
    y_test: list or array
    True label of testing set

    pred_test: list or array
    Predicted label of testing set

    images_test: list or array
    Images in testing set

    model_name: str
    Name of machine learning model

    save_path: str or pathlib.PosixPath
    Path to store the output

    Returns
    -------
    failed_test: dict
    Dictionary that contains the images, true labels and predicted labels of failures
    """
    failed_test = {
        'image': [],
        'truth': [],
        'pred': []
    }

    for i, truth in enumerate(y_test):
        if truth != pred_test[i]:
            failed_test['image'].append(images_test[i])
            failed_test['truth'].append(truth)
            failed_test['pred'].append(pred_test[i])
    if save_path:
        with open(Path(save_path) / (model_name + '_failed.pkl'), 'wb') as f:
            pickle.dump(failed_test, f)

    return failed_test


def add_fbeta(report, y_true, y_pred, fbeta=2):
    """Add fbeta metric into sklearn.metrics.classification_report

    Parameters
    ----------
    report: pandas.core.frame.DataFrame
    Dataframe from sklearn.metrics.classification_report

    y_true: list or array
    True label of testing set

    y_pred: list or array
    Predicted label of testing set

    fbeta: int
    Beta value in f-beta score

    Returns
    -------
    report: pandas.core.frame.DataFrame
    Report that contains fbeta score
    """
    f2_score = list(fbeta_score(y_true, y_pred, average=None, beta=fbeta))
    f2_score.append(fbeta_score(y_true, y_pred, average='micro', beta=fbeta))
    f2_score.append(fbeta_score(y_true, y_pred, average='macro', beta=fbeta))
    f2_score.append(fbeta_score(y_true, y_pred, average='weighted', beta=fbeta))

    report.loc[f'f{fbeta}-score'] = f2_score

    return report


def metrics_report(y_test, pred_test, model_name=None, save_path=None, fbeta=2, le=None, label_names=None):
    """Output the metrics report using sklearn.metrics.classification_report

    Parameters
    ----------
    y_test: list or array
    True label of testing set

    pred_test: list or array
    Predicted label of testing set

    model_name: str
    Name of machine learning model

    save_path: str or pathlib.PosixPath
    Path to store the output

    fbeta: int
    Beta value of F-beta score. Default is 2

    le: sklearn.preprocessing.LabelEncoder
    Label encoder. Otherwise need to configure label_names

    label_names: list
    Names of labels. Each label name's index means y number

    Returns
    -------
    test_report: pandas.core.frame.DataFrame
    Metrics report
    """
    if not label_names:
        label_names = le.classes_
    test_report = pd.DataFrame(metrics.classification_report(y_test, pred_test,
                                                             target_names=label_names, output_dict=True))
    test_report = add_fbeta(test_report, y_test, pred_test, fbeta)

    if save_path:
        test_report.to_pickle(save_path / (model_name + '_report.pkl'))

    return test_report


def draw_cm(defect_name, solar_df=None, y_true=None, y_pred=None, diagonal=True, cbar_size=0.6, linewidth=0.5, save_path=None, model_name=None):
    """Plot the confusion matrix

    Parameters
    ----------
    solar_df: dict or pandas.core.frame.DataFrame
    Dataframe that indicates the true label and predicted label of all solar cells. Column names for true label and
    predicted should be 'y_true' and 'y_pred'.
    If None, y_true and y_pred must be passed

    y_true, y_pred: array or list
    True label and predicted label of all solar cells. It should be passed if solar_df is not passed

    defect_name: dict
    Mapping the value of y into label names.
    E.g. defect_name = {
            0: 'crack',
            1: 'intact',
            2: 'intra',
            3: 'oxygen',
            4: 'solder'
        }

    diagonal: bool
    If true, the true positive number (diagonal of confusion matrix) is shown.

    cbar_size: float
    Size of color bar. Default is 0.6. Check 'sns.heatmap' for more information.

    linewidth: float
    Line width of confusion matrix. Default is 0.5. Check 'sns.heatmap' for more information.

    model_name: str
    Name of machine learning model

    save_path: str or pathlib.PosixPath
    Path to store the output

    Returns
    -------
    solar_pv: pivot
    Pivot table of confusion matrix. Index is ground truth and column is prediction.

    """
    if not solar_df:
        solar_pv = pd.DataFrame({"y_true": y_true, "y_pred":y_pred}).groupby(['y_true', 'y_pred'], as_index=False).size().pivot(
            index='y_true', columns='y_pred', values='size').rename(columns=defect_name, index=defect_name).fillna(0)
    else:
        solar_pv = pd.DataFrame(solar_df)[['y_true', 'y_pred']].groupby(['y_true', 'y_pred'], as_index=False).size().pivot(
            index='y_true', columns='y_pred', values='size').rename(columns=defect_name, index=defect_name).fillna(0)
    
    if not diagonal:
        np.fill_diagonal(solar_pv.values, np.nan)

    ax = sns.heatmap(data=solar_pv, annot=True, fmt='g', cmap='Blues', square=True, cbar_kws={'shrink': cbar_size},
                     linewidths=linewidth)
    plt.xlabel('Pred')
    plt.ylabel('Truth')

    if save_path:
        plt.tight_layout()
        plt.savefig(Path(save_path) / f'{model_name}_cm.png', bbox_inches=0, dpi=600)


def score_compare(reports_dic, score_type, category, hue_order, save_path=None, output=None,
                  legend_pos=(0.5, 0.4), legend_col=2, label_space=1):
    """Compare the metric of different models.

    reports_dic: dict
    Dictionary of metrics report of different models. Metrics report can be output by 'result_analysis.metrics_report'.

    score_type: str
    Type of metric to be compared, e.g. 'recall', 'f1-score'.

    category: list
    List of categories to be compared, e.g. ['crack', 'intact', 'intra', 'oxygen', 'solder']

    hue_order: list
    List of models to be compared.

    save_path: str or pathlib.PosixPath
    Path to store the output

    output: Bool
    If True, return the melted dataframe

    legend_pos, legend_col, label_space:
    Property of the legend. Check matplotlib.pyplot.legend for more information
    """
    scores = pd.DataFrame()
    for model in reports_dic:
        scores[model] = reports_dic[model].loc[score_type, category]

    melt = scores.reset_index().melt('index', var_name='model', value_name=score_type)
    ax = sns.barplot(data=melt, x='index', y=score_type, hue='model', hue_order=hue_order)
    # ax.set_title('Comparison of '+score_type.title())
    ax.set_xlabel('Category')
    ax.set_ylabel(score_type.title())
    plt.legend(loc='upper center', bbox_to_anchor=legend_pos, ncol=legend_col, columnspacing=label_space)
    plt.tight_layout()
    if save_path:
        plt.savefig(Path(save_path) / f'comparison_{score_type}.png', dpi=600)
    if output:
        return melt


# Especially used to collect the labels from YOLO model
def coordinate2inx(coordinate, row=8, col=16, im_shape=[300, 600]):
    """Convert coordinate of top-left corner of bbox into index.
    Index on solar module looks like:
    [[0, 1, 2]
     [3, 4, 5]]

     Parameters
     ----------
     coordinate: list
     [x, y] of top-left corner of bbox

     row, col: int
     number of rows and columns of solar module

     im_shape: list
     Shape of the module image in the form of [height, width]

     Returns
     -------
     inx: int
     Index of the bbox
    """
    inx = col * round(coordinate[1] / (im_shape[0] / row)) + round(coordinate[0] / (im_shape[1] / col))

    return inx


def coordinate2coordinate(coordinate, row=8, col=16, im_shape=[300, 600]):
    """Convert coordinate of top-left corner of bbox into coordinate of solar cell (x_col, y_row).
    Coordinate of solar cell on solar module looks like:
    [[(0,0), (1,0), (2,0)]
     [(0,1), (1,1), (2,1)]]

     Parameters
     ----------
     coordinate: list
     [x, y] of top-left corner of bbox

     row, col: int
     number of rows and columns of solar module

     im_shape: list
     Shape of the module image in the form of [height, width]

     Returns
     -------
     coordinate_cell: int
     (x_col, y_row) of the bbox
    """
    return round(coordinate[0] / (im_shape[1] / col)), round(coordinate[1] / (im_shape[0] / row))


def inx2coordinate(inx, row=8, col=16):
    """Convert inx of bbox into coordinate of solar cell (x_col, y_row).
    Coordinate of solar cell on solar module looks like:
    [[(0,0), (1,0), (2,0)]
     [(0,1), (1,1), (2,1)]]

     Parameters
     ----------
     inx: int
     Index of the bbox

     row, col: int
     number of rows and columns of solar module

     Returns
     -------
     coordinate_cell: int
     (x_col, y_row) of the bbox
    """
    return inx % col, inx // col


def collect_all_cells(ann_path, labels_pred=None, labels_true=None, pred_key="labels_pred", true_key="labels_true", row_col=[8, 16], shape=[300, 600]):
    """Collect the information of all the cells in one module processed by YOLO model

    Parameters
    ----------
    ann_path: str or pathlib.PosixPath
    Path of ann files

    labels_pred, labels_true: dict
    Mapping label names into names of defects.
    E.g. true label:
    label2defect = {
    "crack_bbox": "crack",
    "oxygen_bbox": "oxygen",
    "solder_bbox": "solder",
    "intra_bbox": "intra"
    }

    pred_key, true_key: str
    Key name of labels in the output "cell_info"

    row_col: list
    Number of rows/columns of module

    shape: list
    Shape of the module image in the form of [height, width]

    Returns
    -------
    cell_info: dict
    The dict that contains information of all the cells on one module.
    Information includes name of the module, index of the cell, labels of prediction/truth, x_col and y_row of the cell
    """
    module_name = os.path.splitext(os.path.split(ann_path)[-1])[0]
    #module_name = str(ann_path).split('/')[-1].split('.')[0]

    all_names = np.full(row_col[0] * row_col[1], fill_value=module_name).tolist()
    all_index = list(range(row_col[0] * row_col[1]))
    all_x = np.tile(list(range(row_col[1])), row_col[0]).tolist()
    all_y = np.tile(list(range(row_col[0])), (row_col[1], 1)).T.flatten().tolist()

    cell_info = {
        'module_name': all_names,
        'index': all_index,
        'x': all_x,
        'y': all_y,
    }

    with open(ann_path, 'r') as file:
        data = json.load(file)
    if labels_true:
        all_labels = np.full(row_col[0] * row_col[1], fill_value='intact').tolist()
        if len(data['objects']) > 0:
            for obj in data['objects']:
                classTitle = obj['classTitle']
                if classTitle in labels_true.keys():
                    points = obj['points']['exterior'][0]
                    inx = coordinate2inx(points, row=row_col[0], col=row_col[1], im_shape=shape)
                    all_labels[inx] = labels_true[classTitle]
        cell_info[true_key] = all_labels
    if labels_pred:
        all_second_labels = np.full(row_col[0] * row_col[1], fill_value='intact').tolist()
        if len(data['objects']) > 0:
            for obj in data['objects']:
                classTitle = obj['classTitle']
                if classTitle in labels_pred.keys():
                    points = obj['points']['exterior'][0]
                    inx = coordinate2inx(points, row=row_col[0], col=row_col[1], im_shape=shape)
                    all_second_labels[inx] = labels_pred[classTitle]
        cell_info[pred_key] = all_second_labels
    return cell_info


def collect_defects(ann_path, defects_dic, labels, row_col=[8, 16], shape=[300, 600], mode=1):
    """Only collect the information of defective cells.
    This method is used when applying YOLO model on data without true labels

    Parameters
    ----------
    ann_path: str or pathlib.PosixPath
    Path of ann files

    defects_dic: dict
    E.g.
    defects_info = {
        'module_name': [],
        'index': [],
        'defects': [],
        'x': [],
        'y': [],
        'confidence': []
        }

    labels: dict
    Mapping label names into names of defects.
    E.g. true label:
    label2defect = {
    "crack_bbox": "crack",
    "oxygen_bbox": "oxygen",
    "solder_bbox": "solder",
    "intra_bbox": "intra"
    }

    row_col: list
    Number of rows/columns of module

    shape: list
    Shape of the module image in the form of [height, width]

    mode: int
    If 0, collect the true labels
    If 1, collect the predicted labels

    Returns
    -------
    defects_dic: dict
    The dict that contains information of all defective cells on one module.
    Information includes name of the module, index of the cell, labels of prediction, x_col and y_row of the cell,
    confidence of the model
    """

    module_name = os.path.splitext(os.path.split(ann_path)[-1])[0]
    #module_name = str(ann_path).split('/')[-1].split('.')[0]
    if module_name in defects_dic['module_name']:
        return defects_dic

    with open(ann_path, 'r') as file:
        data = json.load(file)
    if len(data['objects']) > 0:
        for obj in data['objects']:
            classTitle = obj['classTitle']
            points = obj['points']['exterior'][0]
            inx = coordinate2inx(points, row=row_col[0], col=row_col[1], im_shape=shape)
            # x, y = coordinate2coordinate(points, row=row_col[0], col=row_col[1], im_shape=shape)
            x, y = inx2coordinate(inx, row=row_col[0], col=row_col[1])
            defects_dic['module_name'].append(module_name)
            defects_dic['index'].append(inx)
            defects_dic['defects'].append(labels[classTitle])
            defects_dic['x'].append(x)
            defects_dic['y'].append(y)
            if mode == 1:
                confidence = obj['tags'][0]['value']
                defects_dic['confidence'].append(confidence)

    return defects_dic


def get_label_one_module(ann_path, defects_dic, row_col=[8, 16], shape=[300, 600], fill_label=1):
    """Only collect the label(digit label) information of cells in one module

    Parameters
    ----------
    ann_path: str or pathlib.PosixPath
    Path of ann files

    defects_dic: dict
    Mapping the name of labels into digit
    e.g.
    defects_dic_yolo = {
        'crack_bbox_yolo': 0,
        'solder_bbox_yolo': 4,
        'intra_bbox_yolo': 2,
        'oxygen_bbox_yolo': 3

    }

    row_col: list
    Number of rows/columns of module

    shape: list
    Shape of the module image in the form of [height, width]

    fill_label: int
    The digit label of intact cells. Default is 1

    Returns
    -------
    yolo_labels: list
    Digit labels of cells in one module. Index of the list is the index of the cell in the module
    """
    yolo_labels = np.full(row_col[0] * row_col[1], fill_value=fill_label)
    with open(ann_path, 'r') as file:
        data = json.load(file)
    if len(data['objects']) > 0:
        for obj in data['objects']:
            classTitle = obj['classTitle']
            if classTitle in defects_dic:
                points = obj['points']['exterior'][0]
                inx = coordinate2inx(points, row=row_col[0], col=row_col[1], im_shape=shape)
                yolo_labels[inx] = defects_dic[classTitle]

    return yolo_labels


# Check if the module image needs rotation such that the longer edge is closer to the ground
def need_rotate(corners):
    """Check if need rotation. If the left edge is longer than the right one, then rotate.

    Parameters
    ----------
    corners: array
    Sorted coordinates of module corners. The order is top-left, top-right, bottom-left and bottom-right

    Returns
    -------
    bool
    """
    if np.sum((corners[1] - corners[3]) ** 2) < np.sum((corners[0] - corners[2]) ** 2):
        return True
    else:
        return False


def rotate_points(points, width=600, height=300):
    """180 degree rotation of points of bbox

    Parameters
    ----------
    points: list or array
    Coordinates of top-left, top-right, bottom-left and bottom-right points

    width, height: int
    Width/height of perspective transformed module image

    Returns
    -------
    rotated_points: list or array
    180 degree rotation
    """
    return [[width - points[1][0], height - points[1][1]], [width - points[0][0], height - points[0][1]]]


def rotate_ann(origin_dir, ann, save_dir):
    """Rotate the bbox in ann files

    Parameters
    ----------
    origin_dir: str or pathlib.PosixPath
    Directory of original YOLO ann files

    ann: str
    Name of the ann file

    save_dir:
    str or pathlib.PosixPath
    Directory to store rotated YOLO ann files
    """
    with open(Path(origin_dir) / ann, 'r') as file:
        data = json.load(file)
        size = data['size']
    if len(data['objects']) > 0:
        for obj in data['objects']:
            # classTitle = obj['classTitle']
            points = obj['points']['exterior']
            obj['points']['exterior'] = rotate_points(points, width=size['width'], height=size['height'])

    with open(Path(save_dir) / ann, 'w') as file:
        json.dump(data, file)


# position distribution
def plot_heatmap(dataframe, category, defect_col='defects', bar_color='Blues', linewidths=0.5, cbar_size=0.8):
    """Plot the distribution of defective cells

    Parameters
    ----------
    dataframe: Pandas dataframe
    The dataframe of defects.
    Normally use pd.DataFrame(defects_dict) where defects_dict is obtained from "collect_defects()"

    category: str
    The name of the defect to plot, e.g. "crack"

    defect_col: str
    The name of column of the defects

    bar_color: str
    Color of heatmap

    linewidths: float
    Width of separation lines in the heatmap

    cbar_size: float
    Size of colorbar

    Returns
    -------
    defect_pv: pivot table
    pivot table of defect distribution
    """
    defect_df = dataframe.loc[dataframe[defect_col] == category, ['x', 'y']]
    defect_pv = defect_df.groupby(['x', 'y'], as_index=False).size().pivot('y', 'x', 'size')

    ax = sns.heatmap(defect_pv, square=True, cbar_kws={'shrink':cbar_size}, cmap=bar_color, linewidths=linewidths)
    #ax.invert_yaxis()
    ax.xaxis.set_ticks_position('top')
    #ax.set_title(category.split('_')[0].title())

    return defect_pv



import numpy as np
import cv2 as cv
import json
import unittest
import matplotlib.pyplot as plt
from pv_vision.transform_crop.solarmodule import AbstractModule, MaskModule, SplitModule


class TestSolarModule(unittest.TestCase):

    def testSize(self):
        raw_image = cv.imread("img_mask/1-15-A6CB2-39-08_07_09_06_10.png", cv.IMREAD_UNCHANGED)
        raw_module = AbstractModule(raw_image, 8, 16, 0)
        self.assertEqual(raw_image.shape, raw_module.size, "Wrong module size")
        self.assertEqual(raw_image.shape, raw_module.image.shape, "Wrong module size")
        self.assertEqual([8, 16], [raw_module.row, raw_module.col], "Wrong row/col")

    def testResize(self):
        raw_image = cv.imread("img_mask/1-15-A6CB2-39-08_07_09_06_10.png", cv.IMREAD_UNCHANGED)
        raw_module = AbstractModule(raw_image, 8, 16, 0)
        img_resize = raw_module.resize((100, 200))
        self.assertEqual((100, 200), img_resize.shape)
        raw_module.resize((100, 200), in_place=True)
        self.assertEqual((100, 200), raw_module.size)

    def testRotate(self):
        raw_image = cv.imread("img_mask/1-15-A6CB2-39-08_07_09_06_10.png", cv.IMREAD_UNCHANGED)
        raw_module = AbstractModule(raw_image, 8, 16, 0)
        img_rotate = raw_module.rotate(0)
        self.assertEqual((raw_image.shape[1], raw_image.shape[0]), img_rotate.shape)
        raw_module.rotate(0, in_place=True)
        self.assertEqual([16, 8], [raw_module.row, raw_module.col])

    def testCopyChannel(self):
        raw_image = cv.imread("img_mask/1-15-A6CB2-39-08_07_09_06_10.png", cv.IMREAD_UNCHANGED)
        raw_module = AbstractModule(raw_image, 8, 16, 0)
        img_duplicate = raw_module.copy_channel()
        self.assertEqual(3, img_duplicate.shape[-1])
        self.assertEqual(img_duplicate[:, :, 0].tolist(), img_duplicate[:, :, 1].tolist())

    def testSave(self):
        raw_image = cv.imread("img_mask/1-15-A6CB2-39-08_07_09_06_10.png", cv.IMREAD_UNCHANGED)
        raw_module = AbstractModule(raw_image, 8, 16, 0)
        raw_module.resize((100, 200), in_place=True)
        raw_module.save_fig("img_mask/raw_resize.png")
        img_resize = cv.imread('img_mask/raw_resize.png', cv.IMREAD_UNCHANGED)
        self.assertEqual((100, 200), img_resize.shape)


class TestModule(unittest.TestCase):
    def testCase1(self):
        # test maskModule property
        raw_image = cv.imread("img_mask/1-15-A6CB2-39-08_07_09_06_10.png", cv.IMREAD_UNCHANGED)
        raw_module = MaskModule(raw_image, 8, 16, 0)
        self.assertEqual(raw_image.shape, raw_module.size, "Wrong module size")
        self.assertEqual(raw_image.shape, raw_module.image.shape, "Wrong module size")
        self.assertEqual([8, 16], [raw_module.row, raw_module.col], "Wrong row/col")
        self.assertIsNone(raw_module.mask)
        self.assertIsNone(raw_module.corners)
        self.assertIsNone(raw_module.corner_detection_line())
        self.assertIsNone(raw_module.corner_detection_cont())
        self.assertIsNone(raw_module.transform())
        # test load mask
        raw_module.load_mask("img_mask/1-15-A6CB2-39-08_07_09_06_10.png.json")
        self.assertIsNotNone(raw_module.mask)
        # test corner detection method contour
        corners_cnt = raw_module.corner_detection_cont(output=True)
        self.assertIsNotNone(raw_module.corners)
        # test transformed module
        transformed_module = raw_module.transform(600, 300, img_only=False)
        self.assertTrue(transformed_module.is_transformed(14, 6))
        self.assertEqual((300, 600), transformed_module.size, "Wrong module size")
        self.assertEqual([8, 16], [transformed_module.row, transformed_module.col], "Wrong row/col")
        plt.imshow(transformed_module.image, "gray")
        plt.show()
        plt.clf()
        transformed_module.save_fig("img_mask/transform.png")
        # test another corner detection method
        corners_line = raw_module.corner_detection_line(output=True)
        self.assertNotEqual(corners_line.tolist(), corners_cnt.tolist())
        self.assertEqual(corners_line.tolist(), raw_module.corners.tolist())
        transformed_module = raw_module.transform(600, 300, img_only=True)
        self.assertTrue(raw_module.is_transformed(14, 6))
        plt.imshow(transformed_module, "gray")
        plt.show()
        plt.clf()
        # test crop cells
        cells = raw_module.crop_cell(cellsize=32, vl_interval=25, vl_split_size=32,
                                     hl_interval=25, hl_split_size=32, margin=16)
        self.assertEqual(16*8, len(cells))
        plt.imshow(cells[56], "gray")
        plt.show()
        plt.clf()

    def testCase2(self):
        # test SplitModule
        raw_image = cv.imread("img_only/ELH7.jpg", cv.IMREAD_GRAYSCALE)
        module = SplitModule(raw_image, 6, 10, 3)
        cells = module.crop_cell(cellsize=250)
        plt.imshow(cells[14], "gray")
        plt.show()
        plt.clf()
        module.plot_edges()
        plt.show()
        plt.clf()

    def testCase2(self):
        #test MaskModule
        raw_image = cv.imread("img_only/FSEC_8GC14.jpg", cv.IMREAD_GRAYSCALE)
        module = MaskModule(raw_image, 6, 12, 4)
        mask = module.load_mask(thre=0.1, output=True)
        plt.imshow(mask, "gray")
        plt.show()
        plt.clf()
        module.corner_detection_cont(mode=1)
        transformed = module.transform(cellsize=250)
        plt.imshow(transformed, "gray")
        plt.show()
        plt.clf()
        cells = module.crop_cell(cellsize=250, vl_interval=200, vl_split_size=250,
                                 hl_interval=200, hl_split_size=250, margin=100)
        plt.imshow(cells[0], "gray")
        plt.show()
        plt.clf()

if __name__ == '__main__':
    unittest.main()